# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for workflows."""

from collections.abc import Generator
from contextlib import contextmanager
from typing import override

from django.template import Context
from django.test import RequestFactory

from debusine.db.models import WorkRequest
from debusine.server.workflows.models import WorkRequestWorkflowData
from debusine.tasks.models import BaseDynamicTaskData, BaseTaskData
from debusine.web.templatetags.workflows import workflow_runtime_status_small
from debusine.web.views.tables import WorkflowTable

# templatetags loading code will attempt to import tests, debusine.test
# dependencies (such as lxml) may not be available at runtime
try:
    from debusine.test.django import TestCase, preserve_db_task_registry
except ImportError:
    from unittest import TestCase  # type: ignore[assignment]

    @contextmanager
    def preserve_db_task_registry() -> Generator[None]:
        """Type-compatible stub for the real ``preserve_db_task_registry``."""
        yield


class WorkflowRuntimeStatusSmallTests(TestCase):
    """Tests for workflow_runtime_status_small tag."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.workflow = self.playground.create_workflow()

    def test_status_is_completed(self) -> None:
        """Test workflow status is completed."""
        self.workflow.status = WorkRequest.Statuses.COMPLETED
        self.workflow.workflow_runtime_status = (
            WorkRequest.RuntimeStatuses.COMPLETED
        )

        actual = workflow_runtime_status_small(Context(), self.workflow)

        self.assertHTMLEqual(
            actual,
            '<span class="badge text-bg-primary" title="Completed">C</span>',
        )

    def test_status_is_running(self) -> None:
        """Test workflow status is running and runtime status is blocked."""
        self.workflow.status = WorkRequest.Statuses.RUNNING
        self.workflow.workflow_runtime_status = (
            WorkRequest.RuntimeStatuses.BLOCKED
        )

        actual = workflow_runtime_status_small(Context(), self.workflow)

        self.assertHTMLEqual(
            actual,
            '<span class="badge text-bg-info" title="Running">R</span>'
            '<span class="badge text-bg-dark" title="Blocked">B</span>',
        )

    @preserve_db_task_registry()
    def test_runtime_status_is_needs_input(self) -> None:
        """Test workflow runtime status is needs input."""
        # Import this late, since its dependencies (such as lxml) may not be
        # available when Django tries to import everything under
        # templatetags.
        from debusine.server.tasks.tests.helpers import SampleBaseWaitTask

        self.workflow.status = WorkRequest.Statuses.RUNNING
        self.workflow.workflow_runtime_status = (
            WorkRequest.RuntimeStatuses.NEEDS_INPUT
        )

        class WaitNoop(
            SampleBaseWaitTask[BaseTaskData, BaseDynamicTaskData],
        ):
            TASK_VERSION = 1

            @override
            def _execute(self) -> WorkRequest.Results:
                raise NotImplementedError()

        self.workflow.create_child_worker(
            "noop",
        )

        children_needs_input = self.workflow.create_child_wait(
            "waitnoop",
            workflow_data=WorkRequestWorkflowData(needs_input=True),
        )
        self.playground.advance_work_request(
            children_needs_input, mark_running=True
        )

        wr = self.workflow.create_child_wait(
            "waitnoop",
            workflow_data=WorkRequestWorkflowData(needs_input=True),
        )
        self.playground.advance_work_request(wr, mark_running=True)

        self.workflow.create_child_worker("noop").mark_aborted()

        request = RequestFactory().get("/")
        table = WorkflowTable(
            request, WorkRequest.objects.filter(id=self.workflow.id)
        )
        paginator = table.get_paginator(per_page=10)
        context = Context(paginator.get_context_data())

        actual = workflow_runtime_status_small(context, self.workflow)

        self.assertHTMLEqual(
            actual,
            '<span title="Running" class="badge text-bg-info">R</span> '
            f'<a href="{children_needs_input.get_absolute_url()}">'
            '<span title="Needs Input"class="badge text-bg-info">I</span>'
            '</a>',
        )

    def test_runtime_status_is_needs_input_no_work_request(self) -> None:
        """
        Test workflow runtime status is needs input but no work request waiting.

        The workflow was NEEDS_INPUT but no work request is waiting (finished
        before getting the URL).
        """
        self.workflow.status = WorkRequest.Statuses.RUNNING
        self.workflow.workflow_runtime_status = (
            WorkRequest.RuntimeStatuses.NEEDS_INPUT
        )

        request = RequestFactory().get("/")
        table = WorkflowTable(
            request, WorkRequest.objects.filter(id=self.workflow.id)
        )
        paginator = table.get_paginator(per_page=10)
        context = Context(paginator.get_context_data())

        actual = workflow_runtime_status_small(context, self.workflow)

        self.assertHTMLEqual(
            actual,
            '<span title="Running" class="badge text-bg-info">R</span> '
            '<span title="Needs Input" class="badge text-bg-info">I'
            '</span>',
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Views for the server application."""

from debusine.server.views.rest import ProblemResponse

__all__ = [
    "ProblemResponse",
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the Debusine dput-ng uploader."""

import logging
from unittest import mock

import debian.deb822 as deb822
from dput import upload
from dput.exceptions import UploadException

from debusine.artifacts.models import ArtifactCategory
from debusine.artifacts.playground import ArtifactPlayground
from debusine.client.dput_ng.tests.test_utils import (
    ignore_dput_ng_warnings,
    mock_client_config,
    use_local_dput_ng_config,
)
from debusine.client.exceptions import DebusineError
from debusine.client.models import RelationType
from debusine.test import TestCase
from debusine.test.test_utils import create_artifact_response


class TestDebusineUploader(TestCase):
    """Test the uploader."""

    def test_success(self) -> None:
        """Create debusine artifacts for an upload."""
        upload_path = self.create_temporary_directory()
        changes_path = ArtifactPlayground.create_upload(
            upload_path,
            source=True,
            src_name="hello",
            version="1.0-1",
            binary=False,
        ).files["hello_1.0-1_source.changes"]
        base_path = changes_path.parent
        names = [
            "hello_1.0-1.dsc",
            "hello_1.0-1.debian.tar.xz",
            "hello_1.0.orig.tar.gz",
            "hello_1.0-1_source.changes",
        ]

        with (
            use_local_dput_ng_config(include_core=True),
            ignore_dput_ng_warnings(),
            mock.patch("dput.hook.run_hook") as mock_run_hook,
            mock_client_config(
                "https://debusine.example.net/api", "some-token"
            ),
            mock.patch(
                "debusine.client.debusine.Debusine.artifact_create",
                side_effect=[
                    create_artifact_response(
                        id=artifact_id, files_to_upload=files_to_upload
                    )
                    for artifact_id, files_to_upload in ((1, names), (2, []))
                ],
            ) as mock_artifact_create,
            mock.patch(
                "debusine.client.file_uploader.FileUploader._upload_file",
                return_value=True,
            ) as mock_upload_file,
            mock.patch(
                "debusine.client.debusine.Debusine.relation_create"
            ) as mock_relation_create,
            self.assertLogs(logger="dput", level=logging.INFO) as log,
        ):
            upload(str(changes_path), "debusine.debian.net", no_upload_log=True)

        mock_run_hook.assert_has_calls(
            [
                mock.call(hook_name, mock.ANY, mock.ANY, mock.ANY)
                for hook_name in (
                    "debusine-check-workflow",
                    "checksum",
                    "suite-mismatch",
                    "gpg",
                    "debusine-create-workflow",
                )
            ]
        )
        self.assertEqual(
            [
                (args[0][0].category, args[1])
                for args in mock_artifact_create.call_args_list
            ],
            [
                (
                    ArtifactCategory.UPLOAD,
                    {
                        "workspace": "developers",
                        "work_request": None,
                        "expire_at": None,
                    },
                ),
                (
                    ArtifactCategory.SOURCE_PACKAGE,
                    {
                        "workspace": "developers",
                        "work_request": None,
                        "expire_at": None,
                    },
                ),
            ],
        )
        self.assertEqual(mock_upload_file.call_count, 4)
        mock_relation_create.assert_has_calls(
            [
                mock.call(1, 2, RelationType.EXTENDS),
                mock.call(1, 2, RelationType.RELATES_TO),
            ]
        )
        self.assertEqual(
            log.output,
            [
                "INFO:dput:Uploading hello using debusine to "
                "debusine.debian.net "
                "(host: debusine.debian.net; directory: /)",
                "INFO:dput:Not writing upload log upon request",
            ]
            + [f"INFO:dput:Uploading {name}" for name in names]
            + [
                f"INFO:dput:Uploading {base_path}/{name} to "
                f"https://debusine.example.net/api/1.0/artifact/1/files/{name}/"
                for name in names
            ]
            + [
                "INFO:dput:Created artifact: "
                "https://debusine.example.net/debian/developers/artifact/1/",
            ],
        )

    def test_multiple_changes_files(self) -> None:
        """Attempting to upload multiple .changes files is an error."""
        upload_path = self.create_temporary_directory()
        ArtifactPlayground.write_deb822_file(
            deb822.Changes,
            (inner_changes := upload_path / "inner_1_source.changes"),
            [],
            source="inner",
            version="1",
        )
        ArtifactPlayground.write_deb822_file(
            deb822.Changes,
            (outer_changes := upload_path / "outer_1_source.changes"),
            [inner_changes],
            source="outer",
            version="1",
        )

        with (
            use_local_dput_ng_config(include_core=True),
            ignore_dput_ng_warnings(),
            mock.patch("dput.hook.run_hook"),
            mock_client_config(
                "https://debusine.example.net/api", "some-token"
            ),
            mock.patch(
                "debusine.client.debusine.Debusine.upload_artifact",
                return_value=create_artifact_response(id=1),
            ),
            mock.patch("debusine.client.debusine.Debusine.relation_create"),
            self.assertLogs(logger="dput", level=logging.INFO) as log,
            self.assertRaisesRegex(
                RuntimeError, "Cannot upload multiple .changes files"
            ),
        ):
            upload(
                str(outer_changes), "debusine.debian.net", no_upload_log=True
            )

        self.assertEqual(
            log.output,
            [
                "INFO:dput:Uploading outer using debusine to "
                "debusine.debian.net "
                "(host: debusine.debian.net; directory: /)",
                "INFO:dput:Not writing upload log upon request",
                "INFO:dput:Uploading inner_1_source.changes",
                "INFO:dput:Created artifact: "
                "https://debusine.example.net/debian/developers/artifact/1/",
                "INFO:dput:Uploading outer_1_source.changes",
            ],
        )

    def test_upload_artifact_error(self) -> None:
        """
        Exceptions from ``upload_artifact`` are wrapped in an UploadException.

        This is handled cleanly by the ``dput`` command line.
        """
        upload_path = self.create_temporary_directory()
        changes_path = ArtifactPlayground.create_upload(
            upload_path,
            source=True,
            src_name="hello",
            version="1.0-1",
            binary=False,
        ).files["hello_1.0-1_source.changes"]
        error = DebusineError(
            title="Something went wrong",
            detail="Something went badly wrong",
        )

        with (
            use_local_dput_ng_config(include_core=True),
            ignore_dput_ng_warnings(),
            mock.patch("dput.hook.run_hook"),
            mock_client_config(
                "https://debusine.example.net/api", "some-token"
            ),
            mock.patch(
                "debusine.client.debusine.Debusine.upload_artifact",
                side_effect=error,
            ),
            self.assertLogs(logger="dput", level=logging.INFO),
            self.assertRaises(UploadException) as raised,
        ):
            upload(str(changes_path), "debusine.debian.net", no_upload_log=True)

        self.assertIsInstance(raised.exception, UploadException)
        self.assertEqual(str(raised.exception), str(error))
        self.assertEqual(raised.exception.__cause__, error)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the piuparts task support on the worker."""

import datetime as dt
import shlex
import tarfile
import textwrap
from collections.abc import Iterable
from io import BytesIO
from pathlib import Path
from typing import override
from unittest import mock
from unittest.mock import MagicMock, call

from debian.debian_support import Version

import debusine.worker.tags as wtags
from debusine.artifacts import PiupartsArtifact
from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    DebianPiuparts,
    DebianSystemTarball,
    DebianUpload,
    WorkRequestResults,
)
from debusine.client.models import (
    ArtifactResponse,
    FileResponse,
    FilesResponseType,
    RelationType,
)
from debusine.tasks import Piuparts, TaskConfigError
from debusine.tasks.models import (
    BackendType,
    BaseTaskInputArtifact,
    ExtraExternalRepository,
    PiupartsDynamicData,
)
from debusine.tasks.tests.helpers import ExternalTaskHelperMixin
from debusine.test import TestCase
from debusine.test.test_utils import (
    create_artifact_response,
    create_remote_artifact,
    create_system_tarball_data,
)


class PiupartsTests(ExternalTaskHelperMixin[Piuparts], TestCase):
    """Test the Piuparts class."""

    SAMPLE_TASK_DATA = {
        "input": {"binary_artifacts": [421]},
        "build_architecture": "amd64",
        "environment": "debian/match:codename=bookworm",
        "base_tgz": 44,
    }

    _binary_upload_data = DebianUpload(
        type="dpkg",
        changes_fields={
            "Architecture": "amd64",
            "Version": "1.0",
            "Files": [{"name": "hello_1.0_amd64.deb"}],
            "Source": "hello",
        },
    )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.resolver.set(extra_repositories=None)
        self.resolver.set_field_artifact(
            "environment",
            id=1,
            category=ArtifactCategory.SYSTEM_TARBALL,
            data=create_system_tarball_data(),
        )
        self.resolver.set_field_artifact(
            "base_tgz",
            id=44,
            category=ArtifactCategory.SYSTEM_TARBALL,
            data=create_system_tarball_data(),
        )
        self.resolver.set_field_artifacts(
            "input_binary_artifacts",
            BaseTaskInputArtifact(
                artifact_id=421,
                category=ArtifactCategory.UPLOAD,
                data=self._binary_upload_data,
            ),
        )

    def test_configure_fails_with_missing_required_data(
        self,
    ) -> None:
        with self.assertRaises(TaskConfigError):
            self.configure_task(override={"input": {"artifact": 1}})

    def test_configure_environment_is_required(self) -> None:
        """Missing required field "environment": raise exception."""
        with self.assertRaises(TaskConfigError):
            self.configure_task(remove=["environment"])

    def test_configure_base_tgz_is_required(self) -> None:
        """Missing required field "base_tgz": raise exception."""
        with self.assertRaises(TaskConfigError):
            self.configure_task(remove=["base_tgz"])

    def test_get_provided_worker_tags(self) -> None:
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [
                wtags.TASK_PREFIX + "worker:piuparts:version:1",
            ],
        )

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags(
            require_system={
                wtags.BUILD_ARCH_PREFIX + "amd64",
                wtags.EXECUTOR_PREFIX + "unshare",
            }
        )

    def test_compute_dynamic_data(self) -> None:
        """Dynamic data receives relevant artifact IDs."""
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            PiupartsDynamicData(
                environment_id=1,
                input_binary_artifacts_ids=[421],
                base_tgz_id=44,
                subject="hello",
                parameter_summary="hello_1.0 (amd64)",
                runtime_context="sid:amd64",
                configuration_context="sid",
            ),
        )

    def test_get_subject(self) -> None:
        self.assertEqual(self.get_subject(), "hello")

    def test_get_subject_multiple_binaries_same_source(self) -> None:
        self.configure_task(
            override={"input": {"binary_artifacts": [421, 422, 423]}}
        )

        self.resolver.set_field_artifact(
            "environment",
            id=44,
            category=ArtifactCategory.SYSTEM_TARBALL,
            data=create_system_tarball_data(),
        )

        self.resolver.set_field_artifact(
            "base_tgz",
            id=45,
            category=ArtifactCategory.SYSTEM_TARBALL,
            data=create_system_tarball_data(codename="bookworm"),
        )

        binary_1 = DebianUpload(
            type="dpkg",
            changes_fields={
                "Architecture": "amd64",
                "Version": "1.0",
                "Files": [{"name": "linux-cpupower_1.0_amd64.deb"}],
                "Source": "linux",
            },
        )

        binary_2 = DebianUpload(
            type="dpkg",
            changes_fields={
                "Architecture": "amd64",
                "Version": "1.0",
                "Files": [{"name": "linux-image-1.0_amd64.deb"}],
                "Source": "linux",
            },
        )

        binary_3 = DebianUpload(
            type="dpkg",
            changes_fields={
                "Architecture": "amd64",
                "Version": "1.0",
                "Files": [{"name": "hello-1.0.0_amd64.deb"}],
                "Source": "hello",
            },
        )

        self.resolver.set_field_artifacts(
            "input_binary_artifacts",
            BaseTaskInputArtifact(
                artifact_id=421,
                category=ArtifactCategory.UPLOAD,
                data=binary_1,
            ),
            BaseTaskInputArtifact(
                artifact_id=422,
                category=ArtifactCategory.UPLOAD,
                data=binary_2,
            ),
            BaseTaskInputArtifact(
                artifact_id=423,
                category=ArtifactCategory.UPLOAD,
                data=binary_3,
            ),
        )

        self.assertEqual(self.get_subject(), "hello linux")

    def test_get_configuration_context(self) -> None:
        self.assertEqual(self.get_configuration_context(), "sid")

    def test_compute_dynamic_data_base_tgz_constraints(self) -> None:
        """Dynamic data applies environment-like constraints to base_tgz."""
        self.configure_task(
            override={"base_tgz": "debian/match:codename=bookworm"}
        )
        self.resolver.set_field_artifact(
            "environment",
            id=1,
            category=ArtifactCategory.SYSTEM_TARBALL,
            data=create_system_tarball_data(),
        )
        self.resolver.set_field_artifact(
            "base_tgz",
            id=44,
            category=ArtifactCategory.SYSTEM_TARBALL,
            data=create_system_tarball_data(codename="bookworm"),
        )

        self.assertEqual(
            self.compute_dynamic_data(self.task),
            PiupartsDynamicData(
                environment_id=1,
                input_binary_artifacts_ids=[421],
                base_tgz_id=44,
                subject="hello",
                parameter_summary="hello_1.0 (amd64)",
                runtime_context="bookworm:amd64",
                configuration_context="bookworm",
            ),
        )

    def test_compute_subject_parameter_summary_multiple_binaries_same_source(
        self,
    ) -> None:
        self.configure_task(
            override={"input": {"binary_artifacts": [421, 422]}}
        )
        self.resolver.set_field_artifact(
            "environment",
            id=44,
            category=ArtifactCategory.SYSTEM_TARBALL,
            data=create_system_tarball_data(),
        )

        self.resolver.set_field_artifact(
            "base_tgz",
            id=45,
            category=ArtifactCategory.SYSTEM_TARBALL,
            data=create_system_tarball_data(codename="bookworm"),
        )

        binary_1 = DebianUpload(
            type="dpkg",
            changes_fields={
                "Architecture": "amd64",
                "Version": "1.0",
                "Files": [{"name": "linux-cpupower_1.0_amd64.deb"}],
                "Source": "linux",
            },
        )

        binary_2 = DebianUpload(
            type="dpkg",
            changes_fields={
                "Architecture": "amd64",
                "Version": "1.0",
                "Files": [{"name": "linux-image-1.0_amd64.deb"}],
                "Source": "linux",
            },
        )

        self.resolver.set_field_artifacts(
            "input_binary_artifacts",
            BaseTaskInputArtifact(
                artifact_id=421,
                category=ArtifactCategory.UPLOAD,
                data=binary_1,
            ),
            BaseTaskInputArtifact(
                artifact_id=422,
                category=ArtifactCategory.UPLOAD,
                data=binary_2,
            ),
        )

        dynamic_data = self.compute_dynamic_data(self.task)

        self.assertEqual(dynamic_data.subject, "linux")
        self.assertEqual(dynamic_data.parameter_summary, "linux_1.0 (amd64)")

    def test_compute_dynamic_data_multiple_source_packages(self) -> None:
        self.configure_task(
            override={"input": {"binary_artifacts": [421, 422]}}
        )

        self.resolver.set_field_artifact(
            "environment",
            id=44,
            category=ArtifactCategory.SYSTEM_TARBALL,
            data=create_system_tarball_data(),
        )
        self.resolver.set_field_artifact(
            "base_tgz",
            id=45,
            category=ArtifactCategory.SYSTEM_TARBALL,
            data=create_system_tarball_data(codename="bookworm"),
        )
        self.resolver.set_field_artifacts(
            "input_binary_artifacts",
            BaseTaskInputArtifact(
                artifact_id=421,
                category=ArtifactCategory.UPLOAD,
                data=self._binary_upload_data,
            ),
            BaseTaskInputArtifact(
                artifact_id=422,
                category=ArtifactCategory.UPLOAD,
                data=DebianUpload(
                    type="dpkg",
                    changes_fields={
                        "Architecture": "amd64",
                        "Version": "2.0",
                        "Files": [{"name": "bye_1.0-1_amd64.deb"}],
                        "Source": "bye",
                    },
                ),
            ),
        )
        dynamic_data = self.compute_dynamic_data(self.task)

        self.assertEqual(dynamic_data.subject, "bye hello")
        self.assertEqual(dynamic_data.parameter_summary, "bye hello (amd64)")

    def test_acceptable_input_categories(self) -> None:
        self.assertEqual(
            Piuparts.inputs["base_tgz"].categories,  # type: ignore[attr-defined]
            (ArtifactCategory.SYSTEM_TARBALL,),
        )
        self.assertEqual(
            Piuparts.inputs["input_binary_artifacts"].categories,  # type: ignore[attr-defined]
            (
                ArtifactCategory.BINARY_PACKAGE,
                ArtifactCategory.UPLOAD,
            ),
        )

    def test_get_input_artifacts_ids(self) -> None:
        """Test get_input_artifacts_ids."""
        self.task.dynamic_data = None
        self.assertEqual(self.task.get_input_artifacts_ids(), [])

        self.task.dynamic_data = PiupartsDynamicData(
            environment_id=1,
            base_tgz_id=2,
            input_binary_artifacts_ids=[3, 4],
        )
        self.assertEqual(self.task.get_input_artifacts_ids(), [1, 2, 3, 4])

    def test_cmdline_as_root(self) -> None:
        """_cmdline_as_root() return True."""
        self.assertTrue(self.task._cmdline_as_root())

    def create_basetgz_tarfile(
        self,
        destination: Path,
        directories: Iterable[Path] = (),
        symlinks: Iterable[tuple[Path, Path]] = (),
        empty_files: Iterable[Path] = (),
    ) -> None:
        """Create a tar file that would make a useful test for base_tgz."""
        assert destination.suffix == ".xz"
        with tarfile.open(destination, "w:xz") as tar:
            tarinfo = tarfile.TarInfo(name="./foo")
            tarinfo.size = 12
            tar.addfile(tarinfo, BytesIO(b"A tar member"))

            # Add this even if with_dev=False, so that we can test
            # whether the preparation step removes it.
            tarinfo = tarfile.TarInfo(name="./dev/null")
            tarinfo.type = tarfile.CHRTYPE
            tarinfo.devmajor = 1
            tarinfo.devminor = 3
            tar.addfile(tarinfo)

            # We're skipping parent directories, but that's fine, this will
            # never be extracted.
            for path in directories:
                tarinfo = tarfile.TarInfo(name=f".{path}")
                tarinfo.type = tarfile.DIRTYPE
                tar.addfile(tarinfo)

            for src, dst in symlinks:
                tarinfo = tarfile.TarInfo(name=f".{dst}")
                tarinfo.type = tarfile.SYMTYPE
                tarinfo.linkname = str(src)
                tar.addfile(tarinfo)

            for path in empty_files:
                tarinfo = tarfile.TarInfo(name=f".{path}")
                tarinfo.size = 0
                tar.addfile(tarinfo, BytesIO(b""))

    def patch_fetch_artifact_for_basetgz(
        self,
        *,
        with_dev: bool,
        with_resolvconf_symlink: bool = False,
        with_dpkg_available: bool = True,
        codename: str = "bookworm",
        filename: str = "system.tar.xz",
    ) -> MagicMock:
        """Patch self.task.fetch_artifact(), pretend to download basetgz."""
        patcher = mock.patch.object(self.task, "fetch_artifact", autospec=True)
        mocked = patcher.start()

        def _create_basetgz_file(
            artifact_id: int,
            destination: Path,
            default_category: (  # noqa: ARG001, U100
                CollectionCategory | None
            ) = None,
        ) -> ArtifactResponse:
            symlinks: list[tuple[Path, Path]] = []
            empty_files: list[Path] = []
            if with_resolvconf_symlink:
                symlinks.append(
                    (
                        Path("../run/systemd/resolve/stub-resolv.conf"),
                        Path("/etc/resolv.conf"),
                    )
                )
            if with_dpkg_available:
                empty_files.append(Path("/var/lib/dpkg/available"))
            self.create_basetgz_tarfile(
                destination / filename,
                symlinks=symlinks,
                empty_files=empty_files,
            )

            fake_url = "https://not-used"

            data = DebianSystemTarball(
                filename=filename,
                vendor="debian",
                codename=codename,
                components=["main"],
                mirror=fake_url,
                variant=None,
                pkglist={},
                architecture="amd64",
                with_dev=with_dev,
                with_init=False,
            )

            file_model = FileResponse(
                size=100,
                checksums={"sha256": "not-used"},
                type="file",
                url=fake_url,
            )

            files = FilesResponseType({"system.tar.xz": file_model})

            return create_artifact_response(
                id=artifact_id,
                workspace="System",
                category=ArtifactCategory.SYSTEM_TARBALL,
                created_at=dt.datetime.now(tz=dt.UTC),
                data=data.model_dump(),
                download_tar_gz_url=(
                    f"http://localhost/artifact/{artifact_id}?archive=tar.gz"
                ),
                files_to_upload=[],
                files=files,
            )

        mocked.side_effect = _create_basetgz_file

        self.addCleanup(patcher.stop)

        return mocked

    def test_configure_for_execution_from_artifact_id_error_no_debs(
        self,
    ) -> None:
        """configure_for_execution() no .deb-s: return False."""
        download_directory = self.create_temporary_directory()
        (file1 := download_directory / "file1.dsc").write_text("")
        (file2 := download_directory / "file2.changes").write_text("")

        self.assertFalse(self.task.configure_for_execution(download_directory))

        files = sorted([str(file1), str(file2)])
        log_file_contents = (
            self.task.log_directory / "configure_for_execution.log"
        ).read_text()
        self.assertEqual(
            log_file_contents,
            f"There must be at least one *.deb file. Current files: {files}\n",
        )

    def test_configure_for_execution_prepare_executor_base_tgz(self) -> None:
        """configure_for_execution() prepare the executor and base tgz."""
        download_directory = self.create_temporary_directory()
        self.write_deb_file(download_directory / "package1_1.0_all.deb")

        self.patch_prepare_executor_instance()

        with (
            mock.patch.object(self.task, "_base_tar_data", autospec=True),
            mock.patch.object(
                self.task, "_prepare_base_tgz", autospec=True
            ) as prepare_base_tgz_mocked,
            mock.patch.object(
                self.task, "_check_piuparts_version", autospec=True
            ) as check_piuparts_version_mocked,
        ):
            self.assertTrue(
                self.task.configure_for_execution(download_directory)
            )

        run_called_with = [
            call(
                ["apt-get", "update"],
                run_as_root=True,
                check=True,
                stdout=mock.ANY,
                stderr=mock.ANY,
            ),
            call(
                ["apt-get", "--yes", "install", "piuparts"],
                run_as_root=True,
                check=True,
                stdout=mock.ANY,
                stderr=mock.ANY,
            ),
        ]
        assert isinstance(self.task.executor_instance, MagicMock)
        self.task.executor_instance.run.assert_has_calls(run_called_with)

        prepare_base_tgz_mocked.assert_called_with(download_directory)

        check_piuparts_version_mocked.assert_called_once()

    def test_check_piuparts_version_stretch(self) -> None:
        with (
            mock.patch.object(
                self.task, "get_package_version", return_value=Version("0.77")
            ),
            self.assertRaisesRegex(
                TaskConfigError,
                (
                    r"The environment contains a version of piuparts \(0\.77\) "
                    r"that is too old\. Container-based backends require "
                    r"piuparts >= 1\.5\. We recommend a >= trixie environment "
                    r"with an older base_tgz\."
                ),
            ),
        ):
            self.task._check_piuparts_version()

    def test_check_piuparts_version_trixie(self) -> None:
        with mock.patch.object(
            self.task, "get_package_version", return_value=Version("1.6.0")
        ):
            self.task._check_piuparts_version()

    def test_check_piuparts_version_incus_vm(self) -> None:
        self.task = Piuparts(
            {
                **self.SAMPLE_TASK_DATA,
                "backend": BackendType.INCUS_VM,
            }
        )

        with mock.patch.object(
            self.task, "get_package_version"
        ) as mock_get_package_version:
            self.task._check_piuparts_version()

        mock_get_package_version.assert_not_called()

    def test_prepare_base_tgz_raise_assertion_error(self) -> None:
        """_prepare_base_tgz raise AssertionError: executor_instance is None."""
        msg = r"^self\.executor_instance cannot be None$"
        self.assertRaisesRegex(
            AssertionError,
            msg,
            self.task._prepare_base_tgz,
            self.create_temporary_directory(),
        )

    def test_prepare_base_tgz_download_base_tgz_no_processing(self) -> None:
        """
        _prepare_base_tgz does not do any artifact processing (with_dev=False).

        * Download the artifact to the correct directory
        * Set self.task._base_tar to it
        """
        codename = "bullseye"
        filename = "system-image.tar.xz"
        fetch_artifact_mocked = self.patch_fetch_artifact_for_basetgz(
            with_dev=False, codename=codename, filename=filename
        )
        self.set_dynamic_data()

        # Use mock's side effect to set self.task.executor_instance
        self.patch_prepare_executor_instance()
        self.task._prepare_executor_instance()

        download_directory = self.create_temporary_directory()

        self.task._prepare_base_tgz(download_directory)

        destination_dir = download_directory / "base_tar"

        fetch_artifact_mocked.assert_called_with(
            self.task.data.base_tgz, destination_dir
        )
        self.assertEqual(self.task._base_tar, destination_dir / filename)

        with tarfile.open(self.task._base_tar) as tar:
            self.assertEqual(
                tar.getnames(),
                ["./foo", "./dev/null", "./var/lib/dpkg/available"],
            )

    def test_prepare_base_tgz_download_base_tgz_remove_dev_files(self) -> None:
        r"""_prepare_base_tgz download the artifact and remove /dev/\*."""
        self.set_dynamic_data()

        # Use mock's side effect to set self.task.executor_instance
        self.patch_prepare_executor_instance()
        self.task._prepare_executor_instance()

        download_directory = self.create_temporary_directory()
        destination_dir = download_directory / "base_tar"

        fetch_artifact_mocked = self.patch_fetch_artifact_for_basetgz(
            with_dev=True
        )

        system_tar = destination_dir / "system.tar.gz"

        self.task._prepare_base_tgz(download_directory)

        fetch_artifact_mocked.assert_called_with(
            self.task.data.base_tgz, destination_dir
        )

        # Downloaded system.tar.xz, then processed it and wrote the result
        # to system.tar.gz
        self.assertEqual(self.task._base_tar, system_tar)

        with tarfile.open(system_tar, "r:gz") as tar:
            self.assertEqual(
                tar.getnames(), ["./foo", "./var/lib/dpkg/available"]
            )

    def test_prepare_base_tgz_download_base_tgz_resolveconf(self) -> None:
        r"""_prepare_base_tgz download the artifact and install resolv.conf."""
        self.set_dynamic_data()

        # Use mock's side effect to set self.task.executor_instance
        self.patch_prepare_executor_instance()
        self.task._prepare_executor_instance()

        download_directory = self.create_temporary_directory()
        destination_dir = download_directory / "base_tar"

        fetch_artifact_mocked = self.patch_fetch_artifact_for_basetgz(
            with_dev=False, with_resolvconf_symlink=True
        )

        self.task._prepare_base_tgz(download_directory)

        fetch_artifact_mocked.assert_called_with(
            self.task.data.base_tgz, destination_dir
        )

        # Downloaded system.tar.xz, then processed it and wrote the result
        # to system.tar.gz
        system_tar = destination_dir / "system.tar.gz"
        self.assertEqual(self.task._base_tar, system_tar)

        with tarfile.open(system_tar, "r:gz") as tar:
            self.assertEqual(
                tar.getnames(),
                [
                    "./foo",
                    "./dev/null",
                    "./var/lib/dpkg/available",
                    "./etc",
                    "./etc/resolv.conf",
                ],
            )
            self.assertTrue(tar.getmember("./etc/resolv.conf").isreg())

    def test_prepare_base_tgz_download_base_tgz_dpkg_available(self) -> None:
        r"""_prepare_base_tgz download the artifact and touch dpkg available."""
        self.set_dynamic_data()

        # Use mock's side effect to set self.task.executor_instance
        self.patch_prepare_executor_instance()
        self.task._prepare_executor_instance()

        download_directory = self.create_temporary_directory()
        destination_dir = download_directory / "base_tar"

        fetch_artifact_mocked = self.patch_fetch_artifact_for_basetgz(
            with_dev=False, with_dpkg_available=False
        )

        self.task._prepare_base_tgz(download_directory)

        fetch_artifact_mocked.assert_called_with(
            self.task.data.base_tgz, destination_dir
        )

        # Downloaded system.tar.xz, then processed it and wrote the result
        # to system.tar.gz
        system_tar = destination_dir / "system.tar.gz"
        self.assertEqual(self.task._base_tar, system_tar)

        with tarfile.open(system_tar, "r:gz") as tar:
            self.assertEqual(
                tar.getnames(),
                [
                    "./foo",
                    "./dev/null",
                    "./var",
                    "./var/lib",
                    "./var/lib/dpkg",
                    "./var/lib/dpkg/available",
                ],
            )
            self.assertTrue(tar.getmember("./var/lib/dpkg/available").isreg())

    def test_filter_tar_existing_directory(self) -> None:
        directory = self.create_temporary_directory()
        filename = directory / "input.tar.xz"
        self.create_basetgz_tarfile(filename, directories=[Path("/var")])
        output = self.task._filter_tar(
            filename, create_empty_files=[Path("/var/test")]
        )
        with tarfile.open(output, "r:gz") as tar:
            self.assertEqual(
                tar.getnames(),
                [
                    "./foo",
                    "./dev/null",
                    "./var",
                    "./var/test",
                ],
            )

    def prepare_script(self, codename: str, script_name: str) -> str:
        """Build scripts and return the contents of script_name."""
        directory = self.create_temporary_directory()
        with mock.patch.object(self.task, "_base_tar_data", autospec=True):
            assert self.task._base_tar_data
            self.task._base_tar_data.codename = codename
            self.task._prepare_scripts(directory)
        script = directory / script_name
        self.assertTrue(script.exists())
        return script.read_text()

    def test_prepare_scripts_bad_permissions(self) -> None:
        """_prepare_scripts overrides pre_remove_50_find_bad_permissions."""
        self.set_dynamic_data()
        script = self.prepare_script(
            "trixie", "pre_remove_50_find_bad_permissions"
        )
        self.assertIn("Skipping", script)

    def test_prepare_scripts_install_useradd(self) -> None:
        """_prepare_scripts adds a script to install useradd."""
        self.set_dynamic_data()
        script = self.prepare_script("trixie", "post_setup_00_passwd")
        self.assertIn("apt-get -y install passwd", script)

    def prepare_extra_repo_script(self, codename: str) -> str:
        """Build and read the post_chroot_unpack_debusine_extra_repos script."""
        return self.prepare_script(
            codename, "post_chroot_unpack_debusine_extra_repos"
        )

    def test_prepare_scripts_buster(self) -> None:
        """_prepare_scripts with deb822 signed-by apt sources."""
        self.configure_task(
            override={
                "extra_repositories": [
                    {
                        "url": "http://example.net",
                        "suite": "bookworm",
                        "components": ["main"],
                        "signing_key": "KEY A",
                    },
                ]
            }
        )
        self.resolver.field_values["extra_repositories"] = [
            ExtraExternalRepository(
                url="http://example.net",
                suite="bookworm",
                components=["main"],
                signing_key="KEY A",
            )
        ]
        self.set_dynamic_data()
        script = self.prepare_extra_repo_script("buster")
        self.assertIn("mkdir -p /etc/apt/keyrings\n", script)
        key = shlex.quote("KEY A\n")
        self.assertIn(
            f"printf %s {key} > /etc/apt/keyrings/extra_apt_key_0.asc\n",
            script,
        )
        source = shlex.quote(
            textwrap.dedent(
                """\
                Types: deb
                URIs: http://example.net/
                Suites: bookworm
                Components: main
                Signed-By: /etc/apt/keyrings/extra_apt_key_0.asc
                """
            )
        )
        self.assertIn(
            (
                f"printf %s {source} > "
                f"/etc/apt/sources.list.d/extra_repository_0.sources\n"
            ),
            script,
        )

    def test_prepare_scripts_bookworm(self) -> None:
        """_prepare_scripts with deb822 signed-by apt sources."""
        self.configure_task(
            override={
                "extra_repositories": [
                    {
                        "url": "http://example.net",
                        "suite": "bookworm",
                        "components": ["main"],
                        "signing_key": "KEY A",
                    },
                ]
            }
        )
        self.resolver.field_values["extra_repositories"] = [
            ExtraExternalRepository(
                url="http://example.net",
                suite="bookworm",
                components=["main"],
                signing_key="KEY A",
            )
        ]
        self.set_dynamic_data()
        script = self.prepare_extra_repo_script("bookworm")
        self.assertNotIn("/etc/apt/keyrings", script)
        source = shlex.quote(
            textwrap.dedent(
                """\
                Types: deb
                URIs: http://example.net/
                Suites: bookworm
                Components: main
                Signed-By:
                 KEY A
                """
            )
        )
        self.assertIn(
            (
                f"printf %s {source} > "
                f"/etc/apt/sources.list.d/extra_repository_0.sources\n"
            ),
            script,
        )

    def test_prepare_scripts_extra_debusine_repositories(self) -> None:
        """_prepare_scripts with extra repositories hosted by Debusine."""
        signing_key = "\n".join(
            (
                "-----BEGIN PGP PUBLIC KEY BLOCK-----",
                "",
                "ABCDEFGHI",
                "-----END PGP PUBLIC KEY BLOCK-----",
            )
        )
        self.configure_task(
            override={
                "extra_repositories": [{"suite": "trixie-local@debian:suite"}]
            }
        )
        self.resolver.field_values["extra_repositories"] = [
            ExtraExternalRepository(
                url="http://deb.example.com/scope/workspace",
                suite="trixie-local",
                components=["main"],
                signing_key=signing_key,
            )
        ]
        self.set_dynamic_data()
        script = self.prepare_extra_repo_script("trixie")
        self.assertNotIn("/etc/apt/keyrings", script)
        source = shlex.quote(
            textwrap.dedent(
                """\
                Types: deb
                URIs: http://deb.example.com/scope/workspace
                Suites: trixie-local
                Components: main
                Signed-By:
                 -----BEGIN PGP PUBLIC KEY BLOCK-----
                 .
                 ABCDEFGHI
                 -----END PGP PUBLIC KEY BLOCK-----
                """
            )
        )
        self.assertIn(
            (
                f"printf %s {source} > "
                f"/etc/apt/sources.list.d/extra_repository_0.sources\n"
            ),
            script,
        )

    def test_check_directory_for_consistency_errors_none(self) -> None:
        """There are no consistency errors."""
        build_directory = self.create_temporary_directory()
        (build_directory / self.task.CAPTURE_OUTPUT_FILENAME).touch()
        self.assertEqual(
            self.task.check_directory_for_consistency_errors(build_directory),
            [],
        )

    def test_check_directory_for_consistency_errors(self) -> None:
        """There is one consistency error: no piuparts.txt."""
        build_directory = self.create_temporary_directory()

        expected = (
            f"{self.task.CAPTURE_OUTPUT_FILENAME} not in {build_directory}"
        )
        self.assertEqual(
            self.task.check_directory_for_consistency_errors(build_directory),
            [expected],
        )

    def test_upload_artifacts(self) -> None:
        """``upload_artifacts`` uploads an artifact and creates relations."""
        execute_directory = self.create_temporary_directory()
        piuparts_output = execute_directory / self.task.CAPTURE_OUTPUT_FILENAME
        piuparts_output.write_text("piuparts output\n")

        self.task._source_artifacts_ids = [1, 2]
        self.task._deb_versions = {"hello": "1.0", "hello-doc": "1.0"}
        piuparts_version = "1.6.0"
        with mock.patch.object(
            self.task,
            "get_package_version",
            return_value=Version(piuparts_version),
        ):
            self.task._check_piuparts_version()
        codename = "trixie"
        self.task._base_tar_data = create_system_tarball_data(codename=codename)

        debusine_mock = self.mock_debusine()
        workspace_name = "testing"
        uploaded_artifact = create_remote_artifact(
            id=10, workspace=workspace_name
        )
        debusine_mock.upload_artifact.side_effect = [uploaded_artifact]

        self.task.workspace_name = workspace_name
        self.task.work_request_id = work_request_id = 147

        self.task.upload_artifacts(
            execute_directory, execution_result=WorkRequestResults.SUCCESS
        )

        debusine_mock.upload_artifact.assert_called_once_with(
            PiupartsArtifact.create(
                piuparts_output=piuparts_output,
                data=DebianPiuparts(
                    piuparts_version=piuparts_version,
                    binary_packages=self.task._deb_versions,
                    architecture="amd64",
                    distribution=f"debian:{codename}",
                ),
            ),
            workspace=workspace_name,
            work_request=work_request_id,
        )
        expected_relation_create = [
            call(
                uploaded_artifact.id,
                source_artifact_id,
                RelationType.RELATES_TO,
            )
            for source_artifact_id in self.task._source_artifacts_ids
        ]
        self.assertEqual(
            debusine_mock.relation_create.call_count,
            len(expected_relation_create),
        )
        debusine_mock.relation_create.assert_has_calls(expected_relation_create)

    def test_execute(self) -> None:
        """Test full (mocked) execution."""
        self.configure_task(override={"input": {'binary_artifacts': [1, 2]}})
        self.task.dynamic_data = PiupartsDynamicData(
            environment_id=7, input_binary_artifacts_ids=[1, 2], base_tgz_id=42
        )
        debusine_mock = self.mock_debusine()
        self.mock_image_download(debusine_mock)
        self.patch_prepare_executor_instance()
        download_directory = self.create_temporary_directory()

        fetch_artifact_mocked = self.patch_fetch_artifact_for_basetgz(
            with_dev=True, codename="trixie"
        )
        self.assertTrue(self.task.fetch_input(download_directory))
        fetch_artifact_mocked.assert_any_call(1, download_directory)
        fetch_artifact_mocked.assert_any_call(2, download_directory)

        self.write_deb_file(file2 := download_directory / "file2_1.0_amd64.deb")
        self.write_deb_file(file1 := download_directory / "file1_1.0_amd64.deb")
        self.write_deb_file(
            file3 := download_directory / "makedev_2.3.1-97_all.deb"
        )

        with mock.patch.object(
            self.task,
            "get_package_version",
            autospec=True,
            return_value="1.6.0",
        ) as check_piuparts_version_mocked:
            self.assertTrue(
                self.task.configure_for_execution(download_directory)
            )

        check_piuparts_version_mocked.assert_called_once()

        self.assertEqual(self.task._deb_files, [file1, file2, file3])
        self.assertEqual(
            self.task._deb_versions,
            {"file1": "1.0", "file2": "1.0", "makedev": "2.3.1-97"},
        )
        self.assertEqual(
            self.task._cmdline(),
            [
                "piuparts",
                "--keep-sources-list",
                "--allow-database",
                "--warn-on-leftovers-after-purge",
                f"--basetgz={self.task._base_tar}",
                "--scriptsdir=/etc/piuparts/scripts",
                f"--scriptsdir={self.task._scripts_dir}",
                str(file1),
                str(file2),
                str(file3),
            ],
        )

        (download_directory / self.task.CAPTURE_OUTPUT_FILENAME).write_text(
            "piuparts output\n"
        )

        assert self.task.executor is not None
        self.task.executor.system_image = self.fake_system_tarball_artifact()

        self.task.upload_artifacts(
            download_directory, execution_result=WorkRequestResults.SUCCESS
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Task to use debdiff in debusine."""

from dataclasses import dataclass
from itertools import chain
from pathlib import Path
from tempfile import mkdtemp
from typing import override

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine import utils
from debusine.artifacts import DebDiffArtifact
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianBinaryPackage,
    DebianSourcePackage,
    DebianUpload,
    WorkRequestResults,
    get_binary_package_version,
    get_source_package_name,
    get_source_package_version,
)
from debusine.client.models import RelationType
from debusine.tasks import BaseTaskWithExecutor, RunCommandTask, TaskConfigError
from debusine.tasks.inputs import (
    EnvironmentInput,
    MultiInputList,
    SingleInputList,
    Stage,
)
from debusine.tasks.models import DebDiffData, DebDiffDynamicData


@dataclass(frozen=True, kw_only=True)
class ArtifactSummaryBase:
    """Summary of the artifacts."""

    package: str | None
    original_version: str | None
    new_version: str | None


@dataclass(frozen=True, kw_only=True)
class SourceArtifactsSummary(ArtifactSummaryBase):
    """Summary of the source artifacts."""

    ids: list[int]


@dataclass(frozen=True, kw_only=True)
class BinaryArtifactsSummary(ArtifactSummaryBase):
    """Summary of the binary artifacts."""

    ids: list[list[int]]


class DebDiff(
    RunCommandTask[DebDiffData, DebDiffDynamicData],
    BaseTaskWithExecutor[DebDiffData, DebDiffDynamicData],
):
    """Task to use debdiff in debusine."""

    TASK_VERSION = 1

    CAPTURE_OUTPUT_FILENAME = "debdiff.txt"

    environment = EnvironmentInput()
    source_artifacts = SingleInputList(
        field="input.source_artifacts",
        categories=[ArtifactCategory.SOURCE_PACKAGE],
        stage=Stage.PENDING,
    )
    binary_artifacts = MultiInputList(
        field="input.binary_artifacts",
        categories=[ArtifactCategory.UPLOAD, ArtifactCategory.BINARY_PACKAGE],
        stage=Stage.PENDING,
    )

    @override
    def post_init(self) -> None:
        super().post_init()

        # list of directories that fetch_input() downloaded the artifacts
        self._original_dirs: list[Path] = []
        self._new_targets_dirs: list[Path] = []

        # list of files that configure_for_execution() found to pass to debdiff
        self._original_targets: list[Path] = []
        self._new_targets: list[Path] = []

    @staticmethod
    def _validate_binary_original_new_count(
        original_count: int, new_count: int
    ) -> None:
        """
        Validate original_count and new_count for binary artifacts.

        Raise TaskConfigError for invalid combinations.
        """
        if original_count == 0 or new_count == 0:
            raise TaskConfigError(
                f"input.binary_artifacts[0] and input.binary_artifacts[1] "
                f"cannot have zero artifacts "
                f"(got {original_count} and {new_count})"
            )

    def _summarize_source_artifacts(self) -> SourceArtifactsSummary | None:
        source_artifacts = self.source_artifacts
        if not source_artifacts:
            return None

        package = None
        artifacts_ids = []
        for source_artifact in source_artifacts:
            if package is None:
                assert isinstance(source_artifact.data, DebianSourcePackage)
                package = get_source_package_name(source_artifact.data)
            artifacts_ids += source_artifact.resolved_ids()

        assert isinstance(source_artifacts[0].data, DebianSourcePackage)
        assert isinstance(source_artifacts[1].data, DebianSourcePackage)

        original_version = get_source_package_version(source_artifacts[0].data)
        new_version = get_source_package_version(source_artifacts[1].data)

        return SourceArtifactsSummary(
            package=package,
            original_version=original_version,
            new_version=new_version,
            ids=artifacts_ids,
        )

    def _summarize_binary_artifacts(self) -> BinaryArtifactsSummary | None:
        binary_artifacts = self.binary_artifacts
        if not binary_artifacts:
            return None

        binary_artifacts_original = binary_artifacts[0]
        binary_artifacts_new = binary_artifacts[1]

        self._validate_binary_original_new_count(
            len(binary_artifacts_original.artifacts),
            len(binary_artifacts_new.artifacts),
        )

        ids = []
        source_package_names = set()
        binary_artifact_category = None

        for multiple_artifacts in [
            binary_artifacts_original,
            binary_artifacts_new,
        ]:
            ids.append(multiple_artifacts.resolved_ids())

            for artifact in multiple_artifacts.artifacts:
                assert isinstance(
                    artifact.data,
                    (DebianUpload, DebianBinaryPackage),
                )
                if binary_artifact_category is None:
                    binary_artifact_category = artifact.category

                if binary_artifact_category != artifact.category:
                    raise TaskConfigError(
                        f'All binary artifacts must have the same '
                        f'category. Found "{binary_artifact_category}" '
                        f'and "{artifact.category}"'
                    )

                source_package_names.add(get_source_package_name(artifact.data))

            if (
                binary_artifact_category != ArtifactCategory.BINARY_PACKAGE
                and len(multiple_artifacts.artifacts) > 1
            ):
                raise TaskConfigError(
                    "If binary_artifacts source and new contain more than "
                    "one artifact all must be of category "
                    "debian:binary-package. Found: "
                    f"{binary_artifact_category}"
                )

        package = None
        original_version = None
        new_version = None

        if len(source_package_names) == 1:
            (package,) = tuple(source_package_names)
            assert isinstance(
                binary_artifacts_original.artifacts[0].data,
                (DebianUpload, DebianBinaryPackage),
            )
            original_version = get_binary_package_version(
                binary_artifacts_original.artifacts[0].data
            )

            assert isinstance(
                binary_artifacts_new.artifacts[0].data,
                (DebianUpload, DebianBinaryPackage),
            )
            new_version = get_binary_package_version(
                binary_artifacts_new.artifacts[0].data
            )

        return BinaryArtifactsSummary(
            package=package,
            original_version=original_version,
            new_version=new_version,
            ids=ids,
        )

    @override
    def compute_scheduler_tags(
        self,
    ) -> tuple[ttags.TaskProvidedTags, wtags.TaskRequiredTags]:
        provided, required = super().compute_scheduler_tags()

        if (
            summary := self._summarize_source_artifacts()
        ) and summary.package is not None:
            provided.add(
                ttags.ProvenanceProvided.SYSTEM,
                {ttags.SOURCE_PACKAGE_PREFIX + summary.package},
            )

        return provided, required

    @override
    def get_subject(self) -> str | None:
        source_summary = self._summarize_source_artifacts()
        binary_summary = self._summarize_binary_artifacts()

        relevant_summary = source_summary or binary_summary
        assert relevant_summary is not None

        if (package := relevant_summary.package) is not None:
            if source_summary is not None:
                return f"source:{package}"
            else:
                return f"binary:{package}"
        return None

    @override
    def compute_dynamic_data(self) -> DebDiffDynamicData:
        """
        Resolve artifact lookups for this task.

        :subject: ``source:NAME_OF_PACKAGE`` if ``input.source_artifacts``
          is set, otherwise ``binary:NAME_OF_PACKAGE`` if all binary artifacts
          in ``input.binary_artifacts`` originate from the same source package
        :parameter_summary: subject: original_version → new_version (arch)
        """
        source_summary = self._summarize_source_artifacts()
        binary_summary = self._summarize_binary_artifacts()

        relevant_summary = source_summary or binary_summary
        assert relevant_summary is not None

        original_version = relevant_summary.original_version
        new_version = relevant_summary.new_version

        parameter_summary = None
        if (package := relevant_summary.package) is not None:
            parameter_summary = f"{package}: {original_version} → {new_version}"

            if source_summary is not None:
                parameter_summary += " (source)"
            else:
                # TODO: It would probably be better to get the architecture
                # of the input binary artifacts and use that.
                parameter_summary += f" ({self.build_architecture})"

        return DebDiffDynamicData(
            environment_id=self.environment.artifact_id,
            input_source_artifacts_ids=(
                source_summary.ids if source_summary else None
            ),
            input_binary_artifacts_ids=(
                binary_summary.ids if binary_summary else None
            ),
            subject=self.get_subject(),
            parameter_summary=parameter_summary,
        )

    @override
    def _cmdline(self) -> list[str]:
        """
        Build the debdiff command line.

        Use configuration of self.data and self.debdiff_target.
        """
        cmd = [
            "debdiff",
        ]

        cmd.extend(self.data.extra_flags)

        original_targets = map(str, self._original_targets)
        new_targets = map(str, self._new_targets)

        use_from_notation = self._original_targets[0].suffix in (
            ".deb",
            ".udeb",
        )

        if use_from_notation:
            cmd += ["--from", *original_targets, "--to", *new_targets]
        else:
            cmd += [*original_targets, *new_targets]

        return cmd

    def _fetch_artifact_to_temporary(
        self, artifact_id: int, directory: Path
    ) -> Path:
        artifact_directory = Path(
            mkdtemp(dir=directory, prefix=f"{artifact_id}_")
        )
        self.fetch_artifact(artifact_id, artifact_directory)

        return artifact_directory

    @override
    def fetch_input(self, destination: Path) -> bool:
        """Download the required artifacts."""
        assert self.dynamic_data

        if self.dynamic_data.input_source_artifacts_ids:
            assert len(self.dynamic_data.input_source_artifacts_ids) == 2
            self._original_dirs.append(
                self._fetch_artifact_to_temporary(
                    self.dynamic_data.input_source_artifacts_ids[0], destination
                )
            )

            self._new_targets_dirs.append(
                self._fetch_artifact_to_temporary(
                    self.dynamic_data.input_source_artifacts_ids[1], destination
                )
            )

        if self.dynamic_data.input_binary_artifacts_ids:
            assert len(self.dynamic_data.input_binary_artifacts_ids) == 2

            for artifact_id in self.dynamic_data.input_binary_artifacts_ids[0]:
                self._original_dirs.append(
                    self._fetch_artifact_to_temporary(artifact_id, destination)
                )

            for artifact_id in self.dynamic_data.input_binary_artifacts_ids[1]:
                self._new_targets_dirs.append(
                    self._fetch_artifact_to_temporary(artifact_id, destination)
                )

        return True

    @staticmethod
    def _find_file(path: Path) -> Path:
        file = utils.find_file_suffixes(path, [".changes"])

        if file is None:
            file = utils.find_file_suffixes(path, [".deb", ".udeb", ".dsc"])

        # file cannot be None because we ensure the artifact category and
        # artifacts are previously validated
        assert file is not None

        return file

    @override
    def configure_for_execution(
        self,
        download_directory: Path,  # noqa: U100
    ) -> bool:
        """
        Set self._(original|new)_targets to the relevant files.

        Use self._original_dirs and self._new_dirs for finding the specific
        targets for debdiff command.

        :param download_directory: unused
        :return: True if valid files were found
        """
        self._prepare_executor_instance()

        if self.executor_instance is None:
            raise AssertionError("self.executor_instance cannot be None")

        assert self.dynamic_data is not None

        self.executor_instance.run(
            ["apt-get", "update"], run_as_root=True, check=True
        )
        self.executor_instance.run(
            [
                "apt-get",
                "--yes",
                "--no-install-recommends",
                "install",
                "devscripts",
                "diffstat",  # for --diffstat
                "patchutils",  # used to compare .diff.gz files
                "wdiff",  # for --wdiff-source-control
            ],
            run_as_root=True,
            check=True,
        )

        for original_dir in self._original_dirs:
            self._original_targets.append(self._find_file(original_dir))

        for target_dir in self._new_targets_dirs:
            self._new_targets.append(self._find_file(target_dir))

        return True

    @override
    def upload_artifacts(
        self,
        exec_directory: Path,
        *,
        execution_result: WorkRequestResults,  # noqa: U100
    ) -> None:
        """Upload the DebDiffArtifact with the files and relationships."""
        assert self.dynamic_data
        if not self.debusine:
            raise AssertionError("self.debusine not set")

        debdiff_file = exec_directory / self.CAPTURE_OUTPUT_FILENAME

        original = " ".join(path.name for path in self._original_targets)
        new = " ".join(path.name for path in self._new_targets)
        debdiff_artifact = DebDiffArtifact.create(
            debdiff_output=debdiff_file,
            original=original,
            new=new,
            label=f"debdiff {self.dynamic_data.parameter_summary}",
        )

        uploaded = self.debusine.upload_artifact(
            debdiff_artifact,
            workspace=self.workspace_name,
            work_request=self.work_request_id,
        )

        for source_artifact_id in self._source_artifacts_ids:
            self.debusine.relation_create(
                uploaded.id, source_artifact_id, RelationType.RELATES_TO
            )

        if (
            binary_artifacts_ids := self.dynamic_data.input_binary_artifacts_ids
        ) is not None:
            for binary_artifact_id in chain.from_iterable(binary_artifacts_ids):
                self.debusine.relation_create(
                    uploaded.id, binary_artifact_id, RelationType.RELATES_TO
                )

    @override
    def task_result(
        self,
        returncode: int | None,
        execute_directory: Path,  # noqa: U100
    ) -> WorkRequestResults:
        """
        Evaluate task output and return success.

        We don't actually check the output, but use the return code of
        debdiff.
        """
        if returncode in [0, 1]:
            return WorkRequestResults.SUCCESS
        return WorkRequestResults.FAILURE



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Sbuild Workflow."""

import re
from functools import cached_property
from typing import override

from debian.debian_support import DpkgArchTable

from debusine.artifacts import SourcePackage
from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
    DebianSourcePackage,
    get_source_package_name,
)
from debusine.client.models import LookupChildType
from debusine.db.models import Artifact
from debusine.server.workflows import workflow_utils
from debusine.server.workflows.base import Workflow, WorkflowValidationError
from debusine.server.workflows.models import (
    SbuildWorkflowData,
    SbuildWorkflowDynamicData,
    WorkRequestWorkflowData,
)
from debusine.tasks.models import (
    ActionRetryWithDelays,
    ActionUpdateCollectionWithArtifacts,
    ActionUpdateCollectionWithData,
    BackendType,
    InputArtifact,
    InputArtifactSingle,
    SbuildBuildComponent,
    SbuildBuildDepResolver,
    SbuildData,
)


class SbuildWorkflow(Workflow[SbuildWorkflowData, SbuildWorkflowDynamicData]):
    """Build a source package for all architectures of a target distribution."""

    TASK_NAME = "sbuild"

    @override
    def post_init(self) -> None:
        super().post_init()
        if self.data.backend == BackendType.AUTO:
            self.data.backend = BackendType.UNSHARE

    @override
    def compute_dynamic_data(self) -> SbuildWorkflowDynamicData:
        r"""
        Compute dynamic data for this workflow.

        :subject: package name of ``input.source_artifact``
        :parameter_summary: package name of
          ``input.source_artifact``\_"version"
        :source_artifact_id: id or None of the input.source_artifact.
        """
        source_artifact = workflow_utils.source_package(
            self, configuration_key="input.source_artifact"
        )
        source_data = SourcePackage.create_data(source_artifact.data)

        return SbuildWorkflowDynamicData(
            subject=source_data.name,
            parameter_summary=f"{source_data.name}_{source_data.version}",
            source_artifact_id=source_artifact.id,
        )

    @cached_property
    def binary_names(self) -> list[str]:
        """Binary names that may be built."""
        return re.split(
            r"\s*,\s*",
            workflow_utils.source_package_data(
                self, configuration_key="input.source_artifact"
            ).dsc_fields.get("Binary", ""),
        )

    @cached_property
    def architectures(self) -> set[str]:
        """Architectures to build."""
        workflow_arches = set(self.data.architectures)

        package_arches = set(
            workflow_utils.source_package_data(
                self, configuration_key="input.source_artifact"
            )
            .dsc_fields.get("Architecture", "")
            .split()
        )
        if not package_arches:
            raise WorkflowValidationError("package architecture list is empty")

        # Intersect architectures with dsc_fields["Architecture"]
        architectures_to_build = set()
        # Support architecture wildcards (any, <os>-any, any-<cpu>)
        # https://www.debian.org/doc/debian-policy/ch-customized-programs.html#architecture-wildcards
        arch_table = DpkgArchTable.load_arch_table()

        for package_arch in package_arches:
            for workflow_arch in workflow_arches:
                if workflow_arch == "all":
                    # all matches any according to DpkgArchTable, but that's
                    # not useful here.
                    if package_arch == "all":
                        architectures_to_build.add("all")
                elif arch_table.matches_architecture(
                    workflow_arch, package_arch
                ):
                    architectures_to_build.add(workflow_arch)

        if not architectures_to_build:
            raise WorkflowValidationError(
                "None of the workflow architectures are supported"
                " by this package:"
                f" workflow: {', '.join(sorted(workflow_arches))}"
                f" package: {', '.join(sorted(package_arches))}"
            )

        return architectures_to_build

    def _get_build_architecture(self, architecture: str) -> str:
        """
        Get the architecture on which to build for a given architecture.

        `architecture` may be "all".
        """
        # TODO (#358): possibly support XS-Build-Indep-Architecture
        return (
            self.data.arch_all_build_architecture
            if architecture == "all"
            else architecture
        )

    def _get_environment_lookup(self) -> str:
        """
        Build the environment lookup.

        :py:func:`get_environment` will fill in additional constraints.
        """
        vendor, codename = self.data.target_distribution.split(":", 1)

        lookup = f"{vendor}/match:codename={codename}"
        if self.data.environment_variant:
            lookup += f":variant={self.data.environment_variant}"

        return lookup

    def _get_environment(self, architecture: str) -> Artifact:
        """
        Lookup an environment to build on the given architecture.

        This is only used for validation that we have all the environments
        we need; the actual tasks will do their own lookups.
        """
        lookup = self._get_environment_lookup()
        build_architecture = self._get_build_architecture(architecture)

        # set in BaseServerTask.__init__() along with work_request
        try:
            env_id = self.work_request.lookup_environment(
                lookup,
                architecture=build_architecture,
                # TODO: This duplicates logic from
                # BaseTaskWithExecutor.backend.
                backend=(
                    BackendType.UNSHARE
                    if self.data.backend == BackendType.AUTO
                    else self.data.backend
                ),
                default_category=CollectionCategory.ENVIRONMENTS,
                try_variant="sbuild",
            ).artifact.id
        except KeyError:
            raise WorkflowValidationError(
                "environment not found for"
                f" {self.data.target_distribution!r} {architecture!r}"
                f" ({lookup!r})"
            )
        return Artifact.objects.get(id=env_id)

    @override
    def validate_input(self) -> None:
        """Thorough validation of input data."""
        # Validate target_distribution
        if ":" not in self.data.target_distribution:
            raise WorkflowValidationError(
                "target_distribution must be in vendor:codename format"
            )

        # Artifact and architectures are validated by accessing
        # self.architectures
        self.architectures

        _vendor, _codename = self.data.target_distribution.split(":", 1)

        # Make sure we have environments for all architectures we need
        for arch in self.architectures:
            # Attempt to get the environment to validate its lookup
            self._get_environment(arch)

    @override
    def populate(self) -> None:
        """Create sbuild WorkRequests for all architectures."""
        for architecture in sorted(self.architectures):
            self._populate_single(
                architecture, environment=self._get_environment_lookup()
            )

    def _populate_single(self, architecture: str, *, environment: str) -> None:
        """Create a single sbuild WorkRequest."""
        build_architecture = self._get_build_architecture(architecture)
        input_ = self.data.input.model_copy()

        input_.source_artifact = (
            workflow_utils.locate_debian_source_package_lookup(
                self, "input.source_artifact", self.data.input.source_artifact
            )
        )

        source_artifact = self.work_request.lookup_single(
            input_.source_artifact, expect_type=LookupChildType.ARTIFACT
        ).artifact
        assert source_artifact.category == ArtifactCategory.SOURCE_PACKAGE
        debian_source_package = DebianSourcePackage(**source_artifact.data)

        vendor, codename = self.data.target_distribution.split(":", 1)
        extra_repositories = workflow_utils.configure_for_overlay_suite(
            self,
            extra_repositories=self.data.extra_repositories,
            vendor=vendor,
            codename=codename,
            environment=environment,
            backend=self.data.backend,
            architecture=build_architecture,
            try_variant="sbuild",
        )

        build_dep_resolver: SbuildBuildDepResolver | None = None
        aspcud_criteria: str | None = None
        if (vendor, codename) == ("debian", "experimental"):
            # https://salsa.debian.org/dsa-team/mirror/dsa-puppet/-/blob/production/modules/buildd/templates/sbuild.conf.erb
            # TODO: Implement this entirely in task-configuration
            build_dep_resolver = SbuildBuildDepResolver.ASPCUD
            aspcud_criteria = (
                "-count(down),-count(changed,APT-Release:=/experimental/),"
                "-removed,-changed,-new"
            )

        task_data = SbuildData(
            input=input_,
            build_architecture=build_architecture,
            environment=environment,
            backend=self.data.backend,
            build_components=[
                (
                    SbuildBuildComponent.ALL
                    if architecture == "all"
                    else SbuildBuildComponent.ANY
                )
            ],
            build_profiles=self.data.build_profiles,
            extra_repositories=extra_repositories,
            binnmu=self.data.binnmu,
            build_dep_resolver=build_dep_resolver,
            aspcud_criteria=aspcud_criteria,
        )
        wr, _ = self.work_request_ensure_child_worker(
            task_name="sbuild",
            task_data=task_data,
            workflow_data=WorkRequestWorkflowData(
                display_name=f"Build {architecture}",
                step=f"build-{architecture}",
            ),
        )

        self.provides_artifact(
            wr,
            ArtifactCategory.UPLOAD,
            f"{self.data.prefix}build-{architecture}",
            data={
                "binary_names": self.binary_names,
                "architecture": architecture,
                "source_package_name": get_source_package_name(
                    debian_source_package
                ),
            },
        )
        self.provides_artifact(
            wr,
            ArtifactCategory.PACKAGE_BUILD_LOG,
            f"{self.data.prefix}buildlog-{architecture}",
            data={
                "source_package_name": debian_source_package.name,
                "source_package_version": debian_source_package.version,
                "architecture": architecture,
            },
        )

        try:
            build_logs_collection_id = self.lookup_singleton_collection(
                CollectionCategory.PACKAGE_BUILD_LOGS
            ).id
        except KeyError:
            build_logs_collection_id = None

        if build_logs_collection_id is not None:
            source_data = workflow_utils.source_package_data(
                self, configuration_key="input.source_artifact"
            )
            variables = {
                "work_request_id": wr.id,
                "vendor": vendor,
                "codename": codename,
                "architecture": build_architecture,
                "srcpkg_name": source_data.name,
                "srcpkg_version": source_data.version,
            }
            wr.add_event_reaction(
                "on_creation",
                ActionUpdateCollectionWithData(
                    collection=build_logs_collection_id,
                    category=BareDataCategory.PACKAGE_BUILD_LOG,
                    data=variables,
                ),
            )
            wr.add_event_reaction(
                "on_success",
                ActionUpdateCollectionWithArtifacts(
                    collection=build_logs_collection_id,
                    variables=variables,
                    artifact_filters={
                        "category": ArtifactCategory.PACKAGE_BUILD_LOG
                    },
                ),
            )

        if self.data.retry_delays is not None:
            wr.add_event_reaction(
                "on_failure",
                ActionRetryWithDelays(delays=self.data.retry_delays),
            )

        for binary_package_name in self.data.signing_template_names.get(
            architecture, []
        ):
            self.provides_artifact(
                wr,
                ArtifactCategory.BINARY_PACKAGE,
                f"{self.data.prefix}signing-template-{architecture}-"
                f"{binary_package_name}",
                data={
                    "architecture": architecture,
                    "binary_package_name": binary_package_name,
                },
                artifact_filters={
                    "data__deb_fields__Package": binary_package_name
                },
            )

    @override
    def get_input_artifacts(self) -> list[InputArtifact]:
        """Return workflow input artifacts (the source package)."""
        return [
            InputArtifactSingle(
                lookup=self.data.input.source_artifact,
                label="input.source_artifact",
                artifact_id=(
                    None
                    if self.dynamic_data is None
                    else self.dynamic_data.source_artifact_id
                ),
            )
        ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for consumers."""

import asyncio
import datetime as dt
from typing import Any, cast, override
from unittest import mock
from unittest.mock import PropertyMock
from urllib.parse import urlencode

from channels.db import database_sync_to_async
from channels.routing import URLRouter
from channels.testing import WebsocketCommunicator
from django.conf import settings
from django.db import transaction
from django.utils import timezone
from django_prometheus.testutils import assert_metric_diff, save_registry
from rest_framework.fields import DateTimeField

from debusine.db.models import Token, WorkRequest, Worker, Workspace
from debusine.db.playground import scenarios
from debusine.server import consumers
from debusine.server.consumers import WorkerConsumer
from debusine.server.middlewares.token_last_seen_at import (
    TokenLastSeenAtMiddlewareChannels,
)
from debusine.server.routing import websocket_urlpatterns
from debusine.tasks.models import WorkerType
from debusine.test.django import (
    BaseDjangoTestCase,
    TestCase,
    TransactionTestCase,
)
from debusine.test.test_utils import date_time_to_isoformat_rest_framework


class WebSocketURLTests(TestCase):
    """Tests for websocket_urlpatterns."""

    def test_websocket_urlpatterns(self) -> None:
        """Ensure websocket_urlpatterns provides a valid URLRouter config."""
        # https://github.com/python/typeshed/pull/15080
        URLRouter(websocket_urlpatterns)  # type: ignore[arg-type]


class WorkerConsumerMixin(BaseDjangoTestCase):
    """Methods used by different WorkerConsumer tests classes."""

    scenario = scenarios.DefaultScopeUser()

    CONNECTED_PAYLOAD = {
        "type": "websocket.send",
        "text": "connected",
    }

    REQUEST_DYNAMIC_METADATA_PAYLOAD = {
        "type": "websocket.send",
        "text": "request_dynamic_metadata",
    }

    WORK_REQUEST_AVAILABLE_PAYLOAD = {
        "type": "websocket.send",
        "text": "work_request_available",
    }

    @database_sync_to_async
    def acreate_worker(
        self,
        fqdn: str | None = None,
        worker_type: WorkerType = WorkerType.EXTERNAL,
    ) -> Worker:
        """Async version to return a new Worker."""
        return self._create_worker(fqdn=fqdn, worker_type=worker_type)

    def _create_worker(
        self,
        fqdn: str | None = None,
        worker_type: WorkerType = WorkerType.EXTERNAL,
    ) -> Worker:
        """Return a new Worker."""
        worker = cast(TestCase, self).playground.create_worker(
            fqdn=fqdn or "computer.lan", worker_type=worker_type
        )
        worker.set_dynamic_metadata(
            {
                "cpu_cores": 4,
                "sbuild:version": 1,
                "sbuild:available": True,
            }
        )
        return worker

    async def connect(
        self,
        *,
        token_key: str | None = None,
        wait_for_fully_connected: bool = True,
    ) -> WebsocketCommunicator:
        """Return WebsocketCommunicator to api/ws/1.0/worker/connect/."""
        headers = [(b'User-Agent', b'A user agent')]

        if token_key:
            headers.append((b'token', token_key.encode('utf-8')))

        communicator = WebsocketCommunicator(
            TokenLastSeenAtMiddlewareChannels(
                consumers.WorkerConsumer.as_asgi()
            ),
            'api/ws/1.0/worker/connect/',
            headers=headers,
        )

        connected, _ = await communicator.connect()
        self.assertTrue(connected)

        if wait_for_fully_connected:
            self.assertEqual(
                await communicator.receive_json_from(timeout=3),
                self.CONNECTED_PAYLOAD,
            )

        return communicator

    async def connect_to_new_worker(
        self, fqdn: str | None = None
    ) -> tuple[WebsocketCommunicator, Worker]:
        """Return a communicator connected to a new worker."""
        worker = await self.acreate_worker(fqdn=fqdn)
        assert worker.token is not None
        communicator = await self.connect(token_key=worker.token.key)

        # Assert that only a request for dynamic metadata is sent
        await self.assertRequestDynamicMetadata(communicator)
        self.assertTrue(await communicator.receive_nothing())

        await worker.arefresh_from_db()
        self.assertTrue(worker.connected())

        return communicator, worker

    def assertTaskExists(self, task_name: str) -> None:  # pragma: no cover
        """Fail if task_name is not in asyncio.all_tasks()."""
        for task in asyncio.all_tasks():
            if task.get_name() == task_name:
                return

        self.fail(f"Asynchronous task '{task_name}' does not exist")

    def assertTaskNotExists(self, task_name: str) -> None:  # pragma: no cover
        """Fail if task_name is in asyncio.all_tasks()."""
        for task in asyncio.all_tasks():
            if task.get_name() == task_name:
                self.fail(f"Asynchronous task '{task_name}' does exist")

    async def assertRequestDynamicMetadata(
        self, communicator: WebsocketCommunicator
    ) -> None:
        """Assert the next communicator msg is request for dynamic metadata."""
        self.assertEqual(
            await communicator.receive_json_from(),
            self.REQUEST_DYNAMIC_METADATA_PAYLOAD,
        )

    def patch_workerconsumer_REQUEST_DYNAMIC_METADATA_SECONDS_fast(
        self,
    ) -> mock.MagicMock:
        """Patch WorkerConsumer.REQUEST_DYNAMIC_METADATA_SECONDS to be fast."""
        patcher = mock.patch(
            'debusine.server.consumers.'
            'WorkerConsumer.REQUEST_DYNAMIC_METADATA_SECONDS',
            new_callable=PropertyMock,
            side_effect=[0, 3600],
        )

        sleep_until_dynamic_metadata_request_mock = patcher.start()
        self.addCleanup(patcher.stop)

        return sleep_until_dynamic_metadata_request_mock

    async def assert_client_rejected(
        self,
        token_key: str | None,
        reason: str,
        reason_code: str | None = None,
    ) -> None:
        """Assert that the client is rejected with a reason."""
        communicator = await self.connect(
            token_key=token_key, wait_for_fully_connected=False
        )

        msg = {
            'type': 'connection_rejected',
            'reason': reason,
        }
        if reason_code is not None:
            msg["reason_code"] = reason_code

        self.assertEqual(await communicator.receive_json_from(), msg)

        self.assertEqual(
            await communicator.receive_output(), {'type': 'websocket.close'}
        )

        await communicator.disconnect()


class WorkerConsumerTransactionTests(WorkerConsumerMixin, TransactionTestCase):
    """
    Tests for WorkerConsumer class.

    Some tests must be implemented in TransactionTestCase because of a bug
    using the database from a TestCase in async.

    See https://code.djangoproject.com/ticket/30448 and
    https://github.com/django/channels/issues/1091#issuecomment-701361358
    """

    async def test_disconnect(self) -> None:
        """Disconnect marks the Worker as disconnected."""
        communicator, worker = await self.connect_to_new_worker()

        await communicator.disconnect()

        await worker.arefresh_from_db()

        self.assertFalse(worker.connected())

    async def assert_messages_send_on_connect(
        self, token_key: str, expected_msgs: list[dict[str, str]]
    ) -> None:
        """
        Connect authenticating with token_key and expects expected_msgs.

        The order of the received and expected_msgs is not enforced. The
        number of received and expected messages must match.
        """
        communicator = await self.connect(token_key=token_key)

        received_messages = []

        while True:
            try:
                received_messages.append(await communicator.receive_json_from())
            except asyncio.exceptions.TimeoutError:
                break

        self.assertEqual(len(expected_msgs), len(received_messages))

        for message in expected_msgs:
            self.assertIn(message, received_messages)

        try:
            await communicator.disconnect()
        except asyncio.exceptions.CancelledError:  # pragma: no cover
            # asgiref < 3.9.0 swallowed this exception; asgiref 3.9.0
            # re-raises it.  See
            # https://github.com/django/asgiref/issues/518.
            pass

    async def test_connect_valid_token(self) -> None:
        """Connect succeeds and a request for dynamic metadata is received."""
        for worker_type in (WorkerType.EXTERNAL, WorkerType.SIGNING):
            with self.subTest(worker_type=str(worker_type)):
                worker = await self.acreate_worker(worker_type=worker_type)
                assert worker.token is not None

                await self.assert_messages_send_on_connect(
                    worker.token.key, [self.REQUEST_DYNAMIC_METADATA_PAYLOAD]
                )

    async def test_connect_with_celery_token(self) -> None:
        worker = await self.acreate_worker(worker_type=WorkerType.CELERY)
        assert worker.token is not None
        await self.assert_client_rejected(
            worker.token.key,
            "Unexpected worker type, only EXTERNAL and SIGNING supported",
        )

    async def test_connected_send_work_request_available_running(self) -> None:
        """Connect succeeds and consumer send expected two messages."""
        worker = await self.acreate_worker()
        assert worker.token is not None

        await database_sync_to_async(self.playground.create_worker_task)(
            mark_pending=True, worker=worker
        )

        await self.assert_messages_send_on_connect(
            worker.token.key,
            [
                self.REQUEST_DYNAMIC_METADATA_PAYLOAD,
                self.WORK_REQUEST_AVAILABLE_PAYLOAD,
            ],
        )

    async def test_connected_send_work_request_available_pending(self) -> None:
        """Connect succeeds and consumer send expected two messages."""
        worker = await self.acreate_worker()
        assert worker.token is not None

        await database_sync_to_async(self.playground.create_worker_task)(
            mark_running=True, worker=worker
        )

        await self.assert_messages_send_on_connect(
            worker.token.key,
            [
                self.REQUEST_DYNAMIC_METADATA_PAYLOAD,
                self.WORK_REQUEST_AVAILABLE_PAYLOAD,
            ],
        )

    async def test_connected_token_without_worker(self) -> None:
        """
        Connect succeeds and an error message is returned.

        Reason of disconnection: token did not have a worker associated.
        """
        token = await database_sync_to_async(
            self.playground.create_user_token
        )()

        await database_sync_to_async(token.enable)()

        # Assert that the server returns that the token does not exist
        await self.assert_client_rejected(
            token.key,
            "No token header, token not associated to a worker or not enabled",
            reason_code="TOKEN_DISABLED",
        )

    async def test_connected_token_not_enabled(self) -> None:
        """
        Connect succeeds and an error message is returned.

        Reason of disconnection: token was not enabled.
        """

        @transaction.atomic
        def create_token() -> Token:
            token = self.playground.create_worker_token(
                enabled=False, set_worker=False
            )
            Worker.objects.create(registered_at=timezone.now(), token=token)
            return token

        token = await database_sync_to_async(create_token)()

        await self.assert_client_rejected(
            token.key,
            "No token header, token not associated to a worker or not enabled",
            reason_code="TOKEN_DISABLED",
        )

    async def test_disconnect_cancel_request_dynamic_metadata(self) -> None:
        """Disconnect cancels the dynamic metadata refresher."""
        self.assertTaskNotExists('dynamic_metadata_refresher')

        communicator, _ = await self.connect_to_new_worker()

        self.assertTaskExists('dynamic_metadata_refresher')

        await communicator.disconnect()

        self.assertTaskNotExists('dynamic_metadata_refresher')

    async def test_disconnect_leaves_work_request_as_running(self) -> None:
        """Worker disconnect does not interrupt the associated WorkRequest."""
        communicator, worker = await self.connect_to_new_worker()

        work_request: WorkRequest = await database_sync_to_async(
            self.playground.create_worker_task
        )(worker=worker, mark_running=True)

        self.assertEqual(work_request.status, work_request.Statuses.RUNNING)

        await communicator.disconnect()

        await worker.arefresh_from_db()

        await work_request.arefresh_from_db()

        self.assertFalse(worker.connected())

        # Assert that work_request is left as running. The Worker
        # is expected to re-connect and pick it up or to update the
        # status to COMPLETED (via the API).
        #
        # During the execution of a task, debusine-server will send
        # pings to the debusine-worker (via uvicorn defaults). debusine-worker
        # will not be able to respond until the task has finished.
        # This causes debusine server to disconnect the worker during
        # the execution of the task.
        #
        # If the status of the work_request changed from RUNNING to
        # PENDING (for example) when debusine-worker tries to update
        # the status to COMPLETED, debusine-server would reject it because
        # a work-request cannot transition from PENDING to COMPLETED.
        self.assertEqual(work_request.status, work_request.Statuses.RUNNING)

    async def test_disconnect_refreshes_worker_from_db(self) -> None:
        """
        Worker disconnect refreshes worker from DB.

        Otherwise asynchronous changes to the worker (such as editing static
        metadata) may be overwritten.
        """
        communicator, worker = await self.connect_to_new_worker()

        def update_static_metadata(worker: Worker) -> None:
            worker.static_metadata = {"system:architectures": ["amd64", "i386"]}
            worker.save()

        await database_sync_to_async(update_static_metadata)(worker)

        await communicator.disconnect()

        await worker.arefresh_from_db()

        self.assertEqual(
            worker.static_metadata, {"system:architectures": ["amd64", "i386"]}
        )
        self.assertFalse(worker.connected())

    async def test_request_dynamic_metadata_after_connect(self) -> None:
        """Debusine sends request_dynamic_metadata_updated after connect."""
        worker = await self.acreate_worker()
        assert worker.token is not None

        self.patch_workerconsumer_REQUEST_DYNAMIC_METADATA_SECONDS_fast()

        communicator = await self.connect(token_key=worker.token.key)

        # debusine sends a request for dynamic metadata because the worker
        # just connected
        await self.assertRequestDynamicMetadata(communicator)

        # sends another one because in this test the method
        # self.patch_workerconsumer_REQUEST_DYNAMIC_METADATA_SECONDS_fast()
        # was called and there is no waiting time
        await self.assertRequestDynamicMetadata(communicator)

        # Nothing else is received for now (the next request for dynamic
        # metadata would happen in 3600 seconds, as per
        # self.patch_workerconsumer_REQUEST_DYNAMIC_METADATA_SECONDS_fast()
        self.assertTrue(await communicator.receive_nothing())

        await communicator.disconnect()

    async def test_connect_redis_not_available(self) -> None:
        """
        Connect succeeds and an error message is returned.

        For example the Redis server was not available at the time of
        connection.
        """
        patcher = mock.patch(
            'debusine.server.consumers.WorkerConsumer._channel_layer_group_add',
            side_effect=OSError,
            autospec=True,
        )
        mocked = patcher.start()
        self.addCleanup(patcher.stop)

        worker = await database_sync_to_async(self.playground.create_worker)()
        assert worker.token is not None

        await self.assert_client_rejected(
            worker.token.key, "Service unavailable"
        )
        mocked.assert_called()

    async def test_worker_disabled(self) -> None:
        """Worker is disabled and msg disconnection msg sent to the worker."""
        communicator, worker = await self.connect_to_new_worker()

        await communicator.send_input({"type": "worker.disabled"})

        def get_token_hash() -> str:
            assert worker.token is not None
            return worker.token.hash

        hash_ = await database_sync_to_async(get_token_hash)()

        self.assertEqual(
            await communicator.receive_json_from(),
            {
                "type": "connection_closed",
                "reason": f'Token has been disabled (token hash: "{hash_}")',
                "reason_code": "TOKEN_DISABLED",
            },
        )

        await communicator.disconnect()

    async def test_work_request_assigned(self) -> None:
        """Assert it sends message to the worker."""
        communicator, worker = await self.connect_to_new_worker()

        def assign_work_request(worker: Worker) -> None:
            self.playground.create_worker_task(worker=worker)

        await database_sync_to_async(assign_work_request)(worker)

        self.assertEqual(
            await communicator.receive_json_from(),
            {"text": "work_request_available", "type": "websocket.send"},
        )

        await communicator.disconnect()

    async def test_connect_invalid_token(self) -> None:
        """
        Connect succeeds and an error message is returned.

        Reason of disconnection: The token does not exist in the database
        """
        await self.assert_client_rejected(
            "does-not-exist",
            "No token header, token not associated to a worker or not enabled",
            reason_code="TOKEN_DISABLED",
        )

    async def test_metrics(self) -> None:
        registry = save_registry()

        communicator1, _ = await self.connect_to_new_worker(
            fqdn="worker1.example"
        )
        assert_metric_diff(
            registry,
            1,
            "debusine_websocket_connect_requests_total",
            view="ws-connect",
        )
        assert_metric_diff(
            registry,
            1,
            "debusine_active_websocket_connections",
            view="ws-connect",
        )

        communicator2, _ = await self.connect_to_new_worker(
            fqdn="worker2.example"
        )
        assert_metric_diff(
            registry,
            2,
            "debusine_websocket_connect_requests_total",
            view="ws-connect",
        )
        assert_metric_diff(
            registry,
            2,
            "debusine_active_websocket_connections",
            view="ws-connect",
        )

        await self.test_connect_invalid_token()
        assert_metric_diff(
            registry,
            3,
            "debusine_websocket_connect_requests_total",
            view="ws-connect",
        )
        assert_metric_diff(
            registry,
            2,
            "debusine_active_websocket_connections",
            view="ws-connect",
        )

        await communicator1.disconnect()
        assert_metric_diff(
            registry,
            3,
            "debusine_websocket_connect_requests_total",
            view="ws-connect",
        )
        assert_metric_diff(
            registry,
            1,
            "debusine_active_websocket_connections",
            view="ws-connect",
        )

        await communicator2.disconnect()
        assert_metric_diff(
            registry,
            3,
            "debusine_websocket_connect_requests_total",
            view="ws-connect",
        )
        assert_metric_diff(
            registry,
            0,
            "debusine_active_websocket_connections",
            view="ws-connect",
        )


class WorkerConsumerTests(WorkerConsumerMixin, TestCase):
    """Tests for the WorkerConsumer class."""

    async def test_connect_without_token(self) -> None:
        """
        Connect succeeds and an error message is returned.

        Reason of disconnection: missing 'token' header.
        """
        await self.assert_client_rejected(
            None,
            "No token header, token not associated to a worker or not enabled",
            reason_code="TOKEN_DISABLED",
        )

    def test_get_header_value_key_found(self) -> None:
        """WorkerConsumer.get_header_value return the key (str)."""
        worker = WorkerConsumer()

        token = '1fb371ea69dca7b'

        worker.scope = {
            'headers': [(b'token', token.encode('utf-8'))],
        }  # type: ignore[typeddict-item]

        self.assertEqual(worker.get_header_value('token'), token)

    def test_get_header_value_key_not_found(self) -> None:
        """WorkerConsumer.get_header_value return None when key not found."""
        worker = WorkerConsumer()

        worker.scope = {'headers': []}  # type: ignore[typeddict-item]

        self.assertIsNone(worker.get_header_value('token'))

    async def test_disconnect_while_not_connected(self) -> None:
        """
        WorkerConsumer.disconnect returns without exceptions.

        Check that WorkerConsumer.disconnect() does not raise exceptions
        when called with a worker that has not connected.
        """
        worker = WorkerConsumer()
        await worker.disconnect(0)


class WorkRequestCompletedConsumerTests(TransactionTestCase):
    """Tests for OnWorkRequestCompletedConsumer class."""

    CONNECTED_PAYLOAD = {
        "type": "websocket.send",
        "text": "connected",
    }

    @override
    def setUp(self) -> None:
        super().setUp()
        self.token = self.playground.create_user_token()

    async def connect(
        self,
        *,
        token_key: str | None = None,
        query_params: dict[str, Any] = {},
        wait_for_fully_connected: bool = True,
    ) -> WebsocketCommunicator:
        """Connect to the WorkRequestCompleted endpoint."""
        headers = [(b"User-Agent", b"A user agent")]

        if token_key is not None:
            headers.append((b"token", token_key.encode("utf-8")))

        query_params_formatted = urlencode(query_params)

        communicator = WebsocketCommunicator(
            TokenLastSeenAtMiddlewareChannels(
                consumers.OnWorkRequestCompletedConsumer.as_asgi()
            ),
            "api/ws/1.0/work-request/on-completed/?" + query_params_formatted,
            headers=headers,
        )

        connected, _ = await communicator.connect()

        self.assertTrue(connected)

        if wait_for_fully_connected:
            self.assertEqual(
                await communicator.receive_json_from(), self.CONNECTED_PAYLOAD
            )

        return communicator

    async def test_connect(self) -> None:
        """Connect to receive notifications that a Work Request completed."""
        communicator = await self.connect(token_key=self.token.key)

        self.assertTrue(await communicator.receive_nothing())

        await communicator.disconnect()

    async def test_disconnect(self) -> None:
        """Ensure we clean-up, cf. #225."""
        consumer = consumers.OnWorkRequestCompletedConsumer()

        consumer.channel_layer = mock.AsyncMock()
        consumer.channel_name = "some-channel-name"
        consumer._channel_added_to_group = True
        await consumer.disconnect("close_code")

        consumer.channel_layer.group_discard.assert_awaited_once_with(
            "work_request_completed", consumer.channel_name
        )

    async def test_connect_rejected_no_token_header(self) -> None:
        """Connect is rejected: no token was provided."""
        communicator = await self.connect(wait_for_fully_connected=False)

        self.assertEqual(
            await communicator.receive_json_from(),
            {
                "reason": "No token header, token not associated to a user or "
                "not enabled",
                "type": "connection_rejected",
            },
        )

        await communicator.disconnect()

    async def test_connect_rejected_workspace_not_found(self) -> None:
        """Connect is rejected: workspace is not found."""
        communicator = await self.connect(
            token_key=self.token.key,
            query_params={"workspaces": "does-not-exist"},
            wait_for_fully_connected=False,
        )

        self.assertEqual(
            await communicator.receive_json_from(),
            {
                "reason": 'Workspace "does-not-exist" not found',
                "type": "connection_rejected",
            },
        )

        await communicator.disconnect()

    async def test_connect_receive_greater_than_last_modified(self) -> None:
        """
        Consumer filter out WorkRequest that completed_at < completed_at_since.

        There are two work requests:
        -WorkRequest work_request_notified completed_at = now() - 1 minute
        -WorkRequest work_request_to_notify completed at after now()

        Client connects and passes now. Only the work_request_to_notify
        is returned.
        """
        result = WorkRequest.Results.SUCCESS

        now = timezone.now()

        work_request_notified = await database_sync_to_async(
            self.playground.create_worker_task
        )(assign_new_worker=True, mark_running=True, result=result)

        work_request_notified.completed_at = now - dt.timedelta(minutes=1)
        await work_request_notified.asave()

        work_request_to_notify = await database_sync_to_async(
            self.playground.create_worker_task
        )(
            assign_new_worker=True,
            mark_running=True,
            result=result,
        )
        assert work_request_to_notify.completed_at is not None

        communicator = await self.connect(
            token_key=self.token.key, query_params={"completed_at_since": now}
        )

        actual = await communicator.receive_json_from()

        self.assertEqual(
            actual,
            {
                "completed_at": DateTimeField().to_representation(
                    work_request_to_notify.completed_at
                ),
                "result": str(work_request_to_notify.result),
                "text": "work_request_completed",
                "type": "websocket.send",
                "work_request_id": work_request_to_notify.id,
            },
        )

        # Received only one WorkRequest
        self.assertTrue(await communicator.receive_nothing())

        await communicator.disconnect()

    async def test_connect_receive_no_include_last_modified(self) -> None:
        """
        Consumer filter out WorkRequest that completed_at == completed_at_since.

        Only one WorkRequest and completed_at == last_completed. Consumer
        does not return it: the client was already aware of this one.
        """
        result = WorkRequest.Results.SUCCESS

        now = timezone.now()

        await database_sync_to_async(self.playground.create_worker_task)(
            assign_new_worker=True,
            mark_running=True,
            result=result,
            completed_at=now,
        )

        communicator = await self.connect(
            token_key=self.token.key, query_params={"completed_at_since": now}
        )

        self.assertTrue(await communicator.receive_nothing())

        await communicator.disconnect()

    async def test_connect_receive_include_last_modified(self) -> None:
        """
        Consumer return 2 WorkRequests with completed_at == completed_at_since.

        There are two WorkRequests where completed_at == last_completed. The
        client might have processed either one or two of them. The server
        returns both of them to avoid any risk of the client
        processing one of them.

        This case is very unusual: the server marked two WorkRequests
        with the same completed_at to the milliseconds.
        """
        result = WorkRequest.Results.SUCCESS

        now = timezone.now()

        work_request_last_created_1 = await database_sync_to_async(
            self.playground.create_worker_task
        )(assign_new_worker=True, mark_running=True, result=result)

        work_request_last_created_1.completed_at = now
        await work_request_last_created_1.asave()

        work_request_last_created_2 = await database_sync_to_async(
            self.playground.create_worker_task
        )(
            assign_new_worker=True,
            mark_running=True,
            result=result,
            completed_at=now,
        )

        communicator = await self.connect(
            token_key=self.token.key, query_params={"completed_at_since": now}
        )

        actual = await communicator.receive_json_from()

        work_request_completed_at = DateTimeField().to_representation(
            work_request_last_created_1.completed_at
        )

        self.assertEqual(
            actual,
            {
                "completed_at": work_request_completed_at,
                "result": str(work_request_last_created_1.result),
                "text": "work_request_completed",
                "type": "websocket.send",
                "work_request_id": work_request_last_created_1.id,
            },
        )

        actual = await communicator.receive_json_from()
        self.assertEqual(
            actual,
            {
                "completed_at": work_request_completed_at,
                "result": str(work_request_last_created_2.result),
                "text": "work_request_completed",
                "type": "websocket.send",
                "work_request_id": work_request_last_created_2.id,
            },
        )

        self.assertTrue(await communicator.receive_nothing())

        await communicator.disconnect()

    async def create_worker_task(self, workspace: Workspace) -> WorkRequest:
        """Create a work request."""
        work_request = await database_sync_to_async(
            self.playground.create_worker_task
        )(
            workspace=workspace,
            assign_new_worker=True,
            mark_running=True,
            result=WorkRequest.Results.SUCCESS,
        )
        assert isinstance(work_request, WorkRequest)
        return work_request

    async def test_work_request_completed_selected_workspace(self) -> None:
        """Notification is received: from a monitored workspace."""
        workspace_name = "lts"
        workspace = await database_sync_to_async(
            self.playground.create_workspace
        )(name=workspace_name)

        workspace_names = f"{workspace_name},{workspace_name}"
        communicator = await self.connect(
            token_key=self.token.key,
            query_params={"workspaces": workspace_names},
        )

        # Wait that the communicator is fully connected before creating
        # the work request
        self.assertTrue(await communicator.receive_nothing())

        work_request = await self.create_worker_task(workspace)
        assert work_request.completed_at is not None

        actual = await communicator.receive_json_from()
        self.assertEqual(
            actual,
            {
                "type": "websocket.send",
                "text": "work_request_completed",
                "work_request_id": work_request.id,
                "completed_at": date_time_to_isoformat_rest_framework(
                    work_request.completed_at
                ),
                "result": work_request.result,
            },
        )

        await communicator.disconnect()

    async def test_work_request_completed_workspace_not_monitored(self) -> None:
        """Notification not received: from a non-monitored workspace."""
        # Need to exist when connecting
        await database_sync_to_async(self.playground.create_workspace)(
            name=settings.DEBUSINE_DEFAULT_WORKSPACE
        )

        workspace = await database_sync_to_async(
            self.playground.create_workspace
        )(name="lts")

        communicator = await self.connect(
            token_key=self.token.key,
            query_params={"workspaces": settings.DEBUSINE_DEFAULT_WORKSPACE},
        )

        # Wait that the communicator is fully connected before creating
        # the work request
        self.assertTrue(await communicator.receive_nothing())

        result = WorkRequest.Results.SUCCESS

        await database_sync_to_async(self.playground.create_worker_task)(
            workspace=workspace,
            assign_new_worker=True,
            mark_running=True,
            result=result,
        )
        self.assertTrue(await communicator.receive_nothing())

        await communicator.disconnect()

    async def test_work_request_completed_no_workspace_specified(self) -> None:
        """Notification is received: all workspaces are monitored."""
        communicator = await self.connect(token_key=self.token.key)

        result = WorkRequest.Results.SUCCESS

        workspace = await database_sync_to_async(
            self.playground.create_workspace
        )()

        work_request = await self.create_worker_task(workspace)
        assert work_request.completed_at is not None

        actual = await communicator.receive_json_from()
        self.assertEqual(
            actual,
            {
                "type": "websocket.send",
                "text": "work_request_completed",
                "work_request_id": work_request.id,
                "completed_at": date_time_to_isoformat_rest_framework(
                    work_request.completed_at
                ),
                "result": result,
            },
        )

        await communicator.disconnect()

    async def test_metrics(self) -> None:
        registry = save_registry()
        users = [
            await database_sync_to_async(self.playground.create_user)(
                username=username
            )
            for username in ("user1", "user2")
        ]
        tokens = [
            await database_sync_to_async(self.playground.create_user_token)(
                user=user
            )
            for user in users
        ]

        communicator1 = await self.connect(token_key=tokens[0].key)
        assert_metric_diff(
            registry,
            1,
            "debusine_websocket_connect_requests_total",
            view="ws-work-request-on-completed",
        )
        assert_metric_diff(
            registry,
            1,
            "debusine_active_websocket_connections",
            view="ws-work-request-on-completed",
        )

        communicator2 = await self.connect(token_key=tokens[1].key)
        assert_metric_diff(
            registry,
            2,
            "debusine_websocket_connect_requests_total",
            view="ws-work-request-on-completed",
        )
        assert_metric_diff(
            registry,
            2,
            "debusine_active_websocket_connections",
            view="ws-work-request-on-completed",
        )

        await self.test_connect_rejected_no_token_header()
        assert_metric_diff(
            registry,
            3,
            "debusine_websocket_connect_requests_total",
            view="ws-work-request-on-completed",
        )
        assert_metric_diff(
            registry,
            2,
            "debusine_active_websocket_connections",
            view="ws-work-request-on-completed",
        )

        await communicator1.disconnect()
        assert_metric_diff(
            registry,
            3,
            "debusine_websocket_connect_requests_total",
            view="ws-work-request-on-completed",
        )
        assert_metric_diff(
            registry,
            1,
            "debusine_active_websocket_connections",
            view="ws-work-request-on-completed",
        )

        await communicator2.disconnect()
        assert_metric_diff(
            registry,
            3,
            "debusine_websocket_connect_requests_total",
            view="ws-work-request-on-completed",
        )
        assert_metric_diff(
            registry,
            0,
            "debusine_active_websocket_connections",
            view="ws-work-request-on-completed",
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Advisory database locks for Debusine."""

from enum import IntEnum


class LockType(IntEnum):
    """
    The type of a database lock.

    This is a 32-bit value used as the first parameter when taking a lock.
    The second parameter, also a 32-bit value, is available for the part of
    the application that owns each lock type.
    """

    #: Used by :py:class:`debusine.server.tasks.aptmirror`.
    APT_MIRROR = 1

    #: Used by :py:class:`debusine.server.tasks.generate_suite_indexes`.
    GENERATE_SUITE_INDEXES = 2


class LockError(Exception):
    """A lock could not be acquired."""



# Copyright 2020-2023 Enrico Zini <enrico@debian.org>
# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Support for external signon providers.

This is configured by the SIGNON_PROVIDERS variable in django settings.

SIGNON_PROVIDERS is expected to be a sequence of `Provider` instances, one for
each supported signon provider.

Example::

    # Configure salsa.debian.org as identity provider
    SIGNON_PROVIDERS=[
        providers.GitlabProvider(
            name="salsa",
            label="Salsa",
            icon="signon/gitlabian.svg",
            client_id="123client_id",
            client_secret="123client_secret",
            url="https://salsa.debian.org",
            restrict=["email-verified", "group:debian"],
        ),
    ]

The `restrict` setting can be used to restrict local account creation to
remote users that have a specific set of claims. It can be set to a sequence of
strings, each specifying a restriction:

* ``group:name``: the given group name must be present. Multiple groups can be
  OR-ed together by separating them with commas; for example
  ``group:debian,ubuntu`` requires the user to be in one of the ``debian`` or
  ``ubuntu`` groups.
* ``email-verified``: the primary email needs to be reported as verified

All listed restrictions must be met (i.e. they are AND-ed together).
"""

import functools
import json
from collections.abc import Collection
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    import django.http
    import jwcrypto.jwk
    from requests_oauthlib.oauth2_session import _Token

    _Token  # fake usage for vulture

# Note: this module is supposed to be imported from settings.py
#
# Its module-level import list for the case of defining providers should be
# kept accordingly minimal


def get(name: str) -> "Provider":
    """
    Look up a provider by name.

    :param name: name of the provider to look up, matching Provider.name
    :raises ImproperlyConfigured: if no provider with that name has been
                                  defined in settings
    :return: the Provider instance
    """
    from django.conf import settings
    from django.core.exceptions import ImproperlyConfigured

    providers = getattr(settings, "SIGNON_PROVIDERS", None)
    if providers is None:
        raise ImproperlyConfigured(
            f"signon provider {name} requested,"
            " but SIGNON_PROVIDERS is not defined in settings"
        )

    # Lookup provider by name
    for p in providers:
        if p.name == name:
            if not isinstance(p, Provider):
                raise ImproperlyConfigured(
                    f"signon provider {name} requested,"
                    f" but its entry in SIGNON_PROVIDERS setting is not a"
                    f" Provider"
                )
            return p

    raise ImproperlyConfigured(
        f"signon provider {name} requested,"
        " but not found in SIGNON_PROVIDERS setting"
    )


class BoundProvider:
    """
    Request-aware proxy for Provider.

    This class provides provider-specific functionality based on the current
    Django request object.
    """

    def __init__(
        self, provider: "Provider", request: "django.http.HttpRequest"
    ) -> None:
        """
        Construct a BoundProvider from a Provider and a HttpRequest.

        :param provider: provider to bind to a request
        :param request: current Django request
        """
        self.provider = provider
        self.request = request

    def __getattr__(self, name: str) -> Any:
        """Proxy attribute access to the provider definition."""
        return getattr(self.provider, name)

    def logout(self) -> None:
        """Log out an externally authenticated identity."""
        from debusine.server.signon.middleware import RequestSignonProtocol

        assert isinstance(self.request, RequestSignonProtocol)
        self.request.signon.identities.pop(self.provider.name, None)
        self.request.session.pop(f"signon_identity_{self.provider.name}", None)


class Provider:
    """Information about a signon identity provider."""

    #: Identifier to reference the provider in code and configuration
    name: str
    #: User-visible description
    label: str
    #: Optional user-visible icon, resolved via ``{% static %}`` in templates
    icon: str | None
    #: Freeform options used to configure behaviour for Signon subclasses
    options: dict[str, Any]
    #: Class used to create a request-bound version
    bound_class: type["BoundProvider"] = BoundProvider

    def __init__(
        self,
        name: str,
        label: str,
        *,
        icon: str | None = None,
        options: dict[str, Any] | None = None,
    ) -> None:
        """
        Define an external authentication provider.

        Provider implementation subclasses can definer further keyword
        arguments.
        """
        self.name = name
        self.label = label
        self.icon = icon
        self.options: dict[str, Any] = options or {}

    def bind(self, request: "django.http.HttpRequest") -> "BoundProvider":
        """Create a BoundProvider for this session."""
        return self.bound_class(self, request)


class OIDCValidationError(Exception):
    """Exception raised when OIDC authentication fails validation."""


class BoundOIDCProvider(BoundProvider):
    """Bound version of the OpenID Connect provider."""

    provider: "OIDCProvider"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Construct a BoundProvider for OpenID Connect."""
        super().__init__(*args, **kwargs)
        from django.urls import reverse
        from requests_oauthlib import OAuth2Session

        self.oauth = OAuth2Session(
            self.provider.client_id,
            scope=self.provider.scope,
            redirect_uri=self.request.build_absolute_uri(
                # FIXME: the URL name is currently dependent on the way this
                # module is deployed. This needs rethinking if we want to
                # publish this as a reusable module
                reverse("signon:oidc_callback", args=(self.name,))
            ),
        )
        self.tokens: _Token | None = None
        self.id_token_claims: dict[str, Any] | None = None
        self.options: Collection[str] | None = None

    def get_authorization_url(self, *options: str) -> str:
        """Return an authorization URL for this provider."""
        url, state = self.oauth.authorization_url(self.provider.url_authorize)
        self.request.session[f"signon_state_{self.provider.name}"] = state
        self.request.session[f"signon_state_{self.provider.name}_options"] = (
            options
        )
        assert isinstance(url, str)
        return url

    @functools.cached_property
    def keyset(self) -> "jwcrypto.jwk.JWKSet":
        """Load crypto keys from settings."""
        # TODO: better key caching can be considered in the future.
        #
        # This caches the server keys forever in process memory.
        #
        # If usage scales up, this can be changed to introduce a way to cache
        # the keys into persistent storage, to avoid one call to url_jwks per
        # process.
        #
        # Also, if the way this is deployed means that processes are very long
        # lived, more attention should be paid to deal with key rotation.
        #
        # See https://openid.net/specs/openid-connect-core-1_0.html#RotateSigKeys  # noqa: E501
        import jwcrypto.jwk

        key_response = self.oauth.get(self.provider.url_jwks)
        key_response.raise_for_status()
        return jwcrypto.jwk.JWKSet.from_json(key_response.text)

    def load_tokens(self) -> None:
        """Fetch and validate access_token and id_token from OIDC provider."""
        import jwcrypto.jwt
        from django.utils.crypto import constant_time_compare

        expected_state = self.request.session.pop(
            f"signon_state_{self.provider.name}", None
        )
        options = self.request.session.pop(
            f"signon_state_{self.provider.name}_options", None
        )

        if expected_state is None:
            raise OIDCValidationError(
                "Request state mismatch: expected state not found in session"
            )
        if options is None:
            raise OIDCValidationError(
                "Request state mismatch: options not found in session"
            )

        remote_state = self.request.GET.get("state")
        if remote_state is None:
            raise OIDCValidationError(
                "Request state mismatch: state not found in remote response"
            )

        if not constant_time_compare(remote_state, expected_state):
            raise OIDCValidationError(
                "Request state mismatch:"
                f" remote: {remote_state!r},"
                f" expected: {expected_state!r}"
            )

        tokens = self.oauth.fetch_token(
            self.url_token,
            authorization_response=self.request.build_absolute_uri(),
            client_secret=self.client_secret,
        )

        # See https://openid.net/specs/openid-connect-core-1_0.html#IDTokenValidation  # noqa: E501

        id_token = tokens["id_token"]

        tok = jwcrypto.jwt.JWT(key=self.keyset, jwt=id_token)
        id_token_claims = json.loads(tok.claims)

        if not constant_time_compare(
            id_token_claims["iss"], self.provider.url_issuer
        ):
            raise OIDCValidationError(
                f"Issuer mismatch: remote: {id_token_claims['iss']!r},"
                f" expected: {self.provider.url_issuer!r}"
            )

        if not constant_time_compare(
            id_token_claims["aud"], self.provider.client_id
        ):
            raise OIDCValidationError(
                f"Audience mismatch: remote: {id_token_claims['aud']!r},"
                f" expected: {self.provider.client_id!r}"
            )

        # Note: the 'exp' claim is checked by default by JWT

        # TODO: potential extra checks to be implemented if needed:
        #
        # * assert tok["iat"] to be not too old
        #
        # OpenID Connect Core 1.0 states:
        #
        # > The iat Claim can be used to reject tokens that were issued too far
        # > away from the current time, limiting the amount of time that nonces
        # > need to be stored to prevent attacks. The acceptable range is
        # > Client specific.
        #
        # * honor tok["auth_time"]
        #
        # OpenID Connect Core 1.0 states:
        #
        # > If the auth_time Claim was requested, either through a specific
        # > request for this Claim or by using the max_age parameter, the
        # > Client SHOULD check the auth_time Claim value and request
        # > re-authentication if it determines too much time has elapsed since
        # > the last End-User authentication.

        self.tokens = tokens
        self.id_token_claims = id_token_claims
        self.options = options


class OIDCProvider(Provider):
    """OpenID Connect identity provider."""

    bound_class = BoundOIDCProvider

    def __init__(
        self,
        *args: Any,
        client_id: str,
        client_secret: str,
        url_issuer: str,
        url_authorize: str,
        url_token: str,
        url_userinfo: str,
        url_jwks: str,
        scope: str | Collection[str],
        restrict: Collection[str] = ("email-verified",),
        group_claim_key: str = "groups",
        **kwargs: Any,
    ) -> None:
        """
        Define an OpenID Connect provider.

        :param client_id: client identifier configured in the authentication
            server
        :param client_secret: client_secret provided by the authentication
            server
        :param url_issuer: URL identifying the authentication server
        :param url_authorize: OIDC authorization endpoint
        :param url_token: OIDC token endpoint
        :param url_userinfo: OIDC userinfo endpoint
        :param url_jwks: OIDC jwks_uri to retrieve the authentication server
            signing keys
        :param restrict: list of restrictions for mapping remote accounts to
                         local ones

        See https://openid.net/specs/openid-connect-core-1_0.html for details
        """
        super().__init__(*args, **kwargs)
        self.client_id = client_id
        self.client_secret = client_secret
        self.url_issuer = url_issuer
        self.url_authorize = url_authorize
        self.url_token = url_token
        self.url_userinfo = url_userinfo
        self.url_jwks = url_jwks
        self.group_claim_key = group_claim_key
        self.scope: list[str]
        if isinstance(scope, str):
            self.scope = [scope]
        else:
            self.scope = list(scope)
        self.restrict: Collection[str] = tuple(restrict)


class AzureEntraProvider(OIDCProvider):
    """Microsoft Azure Entra OIDC identity provider."""

    def __init__(self, *args: Any, tenant_id: str, **kwargs: Any) -> None:
        """
        Define a Azure Entra OIDC Connect provider.

        :param tenant_id: Directory Tenant ID in Azure. It will be used to
            automatically generate all ``url_*`` arguments for OIDCProvider
        """
        kwargs.setdefault("scope", ["openid", "email", "profile"])
        # Entra does not support email_verified
        kwargs.setdefault("restrict", [])
        url_base = f"https://login.microsoftonline.com/{tenant_id}"
        kwargs["url_issuer"] = f"{url_base}/v2.0"
        kwargs["url_authorize"] = f"{url_base}/oauth2/v2.0/authorize"
        kwargs["url_token"] = f"{url_base}/oauth2/v2.0/token"
        kwargs["url_userinfo"] = "https://graph.microsoft.com/oidc/userinfo"
        kwargs["url_jwks"] = f"{url_base}/discovery/v2.0/keys"
        super().__init__(*args, **kwargs)


class GitlabProvider(OIDCProvider):
    """Gitlab OIDC identity provider."""

    def __init__(self, *args: Any, url: str, **kwargs: Any) -> None:
        """
        Define a GitLab-based OIDC Connect provider.

        :param url: URL to the root of the GitLab server. It will be used to
            automatically generate all ``url_*`` arguments for OIDCProvider
        """
        kwargs.setdefault("scope", "openid")
        kwargs["url_issuer"] = url
        kwargs["url_authorize"] = f"{url}/oauth/authorize"
        kwargs["url_token"] = f"{url}/oauth/token"
        kwargs["url_userinfo"] = f"{url}/oauth/userinfo"
        kwargs["url_jwks"] = f"{url}/oauth/discovery/keys"
        kwargs["group_claim_key"] = "groups_direct"
        super().__init__(*args, **kwargs)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for model permissions."""

import re
from collections.abc import Collection, Generator
from typing import Any, ClassVar, Protocol, Self, override

from django.core.exceptions import PermissionDenied
from django.db import models

from debusine.db.context import context
from debusine.db.models import Group, Scope, Token, User, Workspace
from debusine.db.models.permissions import (
    Allow,
    enforce,
    format_permission_check_error,
    get_resource_scope,
    get_resource_workspace,
    permission_check,
    permission_filter,
    resolve_role,
    resolve_roles_list,
)
from debusine.db.permissioncontext import PermissionContext, PermissionUser
from debusine.test.django import (
    AllowAll,
    DenyAll,
    TestCase,
    override_permission,
)


class PredicateWrapper(Protocol):
    """Structural typing for check/filter predicate testers."""

    last_user: PermissionUser | None
    called: bool

    def test(self, pc: PermissionContext) -> bool: ...


class PredicateTester:
    """Encapsulate tests for check and filter predicates."""

    def __init__(
        self,
        test_case: TestCase,
        pc: PermissionContext,
        tester: PredicateWrapper,
    ) -> None:
        """Test calling tester with user."""
        self.test_case = test_case
        self.pc = pc
        self.tester = tester

    @override
    def __repr__(self) -> str:
        """Representation for debugging, shown by subTest."""
        return f"{self.pc=!r}, {self.tester=!r}"

    def test(self) -> bool:
        """Invoke the wrapper's test method."""
        return self.tester.test(self.pc)

    def assert_forced_false(self) -> None:
        """The predicate failed without calling its body."""
        self.test_case.assertFalse(self.test())
        self.test_case.assertFalse(self.tester.called)

    def assert_called(self) -> None:
        """The predicate body has been called with the user passed to it."""
        self.test_case.assertFalse(self.test())
        self.test_case.assertTrue(self.tester.called)
        self.test_case.assertEqual(self.tester.last_user, self.pc.user)

    def assert_passes_with_checks_disabled(self) -> None:
        """The predicate always succeeds with permission checks disabled."""
        orig = self.pc
        try:
            self.pc = self.pc.with_permission_checks_disabled()
            self.test_case.assertTrue(self.test())
            self.test_case.assertFalse(self.tester.called)
        finally:
            self.pc = orig


def mock_pred_check(
    test_case: TestCase,
    *,
    pc: PermissionContext,
    work_request: Allow,
    anonymous: Allow,
) -> PredicateTester:
    """Create a tester for check predicates."""

    class CheckContainer:
        """Mock resource used to test the permission_check decorator."""

        def __init__(self) -> None:
            """Make space to store a trace of checks."""
            self.last_user: PermissionUser | None = None
            self.called = False

        @override
        def __repr__(self) -> str:
            return f"permission_check({work_request=!r}, {anonymous=!r}, {pc=})"

        # permission_check wants a Model, but it's easier to have a mock class
        # that is not a model to avoid having to deal with Django's model
        # registry. Override the check instead
        @permission_check(  # type: ignore[type-var]
            "{user} cannot pred {resource}",
            work_request=work_request,
            anonymous=anonymous,
        )
        def _predicate(self, pc: PermissionContext) -> bool:
            self.last_user = pc.user
            self.called = True
            return False

        def test(self, pc: PermissionContext) -> bool:
            self.called = False
            return self._predicate(pc)

    return PredicateTester(test_case, pc, CheckContainer())


def mock_pred_filter(
    test_case: TestCase,
    *,
    pc: PermissionContext,
    work_request: Allow,
    anonymous: Allow,
) -> PredicateTester:
    """Create a tester for filter predicates."""

    class FilterContainer(models.QuerySet[User]):
        """Mock resource used to test the permission_filter decorator."""

        def __init__(self, *args: Any, **kwargs: Any) -> None:
            """Make space to store a trace of checks."""
            super().__init__(*args, **kwargs)
            self.last_user: PermissionUser | None = None
            self.called = False

        @override
        def __repr__(self) -> str:
            return (
                f"permission_filter({work_request=!r}, {anonymous=!r}, {pc=})"
            )

        @permission_filter(work_request=work_request, anonymous=anonymous)
        def _predicate(self, pc: PermissionContext) -> Self:
            self.last_user = pc.user
            self.called = True
            return self.none()

        def test(self, pc: PermissionContext) -> bool:
            self.called = False
            return self._predicate(pc).count() > 0

    return PredicateTester(test_case, pc, FilterContainer(User))


class PermissionsTests(TestCase):
    """Tests for permission infrastructure."""

    scope: ClassVar[Scope]
    scope1: ClassVar[Scope]
    user: ClassVar[User]
    user1: ClassVar[User]
    group: ClassVar[Group]
    work_request_token: ClassVar[Token]
    work_request_pc: ClassVar[PermissionContext]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.scope = cls.playground.get_default_scope()
        cls.scope1 = cls.playground.get_or_create_scope("scope")
        cls.user = cls.playground.get_default_user()
        cls.user1 = cls.playground.create_user("user1")
        cls.group = cls.playground.create_group_role(
            cls.scope, Scope.Roles.OWNER
        )
        cls.work_request_token = cls.playground.create_work_request_token()
        cls.work_request_pc = PermissionContext.for_work_request(
            token=cls.work_request_token
        )

    def test_resolve_role(self) -> None:
        """Test passing a valid role type."""
        for val, result in (
            ("owner", Workspace.Roles.OWNER),
            (Workspace.Roles.OWNER, Workspace.Roles.OWNER),
        ):
            with self.subTest(val=repr(val)):
                self.assertEqual(resolve_role(Workspace, val), result)

    def test_resolve_role_invalid_type(self) -> None:
        """Test passing an invalid role type."""
        with self.assertRaisesRegex(
            ValueError, r"42 is not a valid ScopeRoles"
        ):
            resolve_role(Scope, 42)  # type: ignore[arg-type]

    def test_resolve_role_invalid_enum(self) -> None:
        """Test passing an invalid role enum."""
        with self.assertRaisesRegex(
            TypeError,
            re.escape(
                f"{Workspace.Roles.OWNER!r} cannot be converted to"
                " <enum 'ScopeRoles'>"
            ),
        ):
            resolve_role(Scope, Workspace.Roles.OWNER)

    def test_resolve_role_invalid_name(self) -> None:
        """Test passing an invalid role name."""
        with self.assertRaisesRegex(
            ValueError, r"'does_not_exist' is not a valid ScopeRoles"
        ):
            resolve_role(Scope, "does_not_exist")

    def test_resolve_roles_list(self) -> None:
        """Test resolve_roles_list function."""
        for roles_cls, arg, expected in (
            (Scope, "owner", [Scope.Roles.OWNER]),
            (Scope, Scope.Roles.OWNER, [Scope.Roles.OWNER]),
            (
                Workspace,
                ["owner", Workspace.Roles.OWNER],
                [Workspace.Roles.OWNER, Workspace.Roles.OWNER],
            ),
        ):
            with self.subTest(roles_cls=repr(roles_cls), arg=repr(arg)):
                self.assertEqual(resolve_roles_list(roles_cls, arg), expected)

    def test_permission_error_template(self) -> None:
        """Test accessing the permission error template."""
        res = mock_pred_check(
            self,
            pc=PermissionContext.for_ui_anonymous(),
            work_request=Allow.PASS,
            anonymous=Allow.PASS,
        )
        resource = res.tester
        resource_class = resource.__class__
        cls_predicate = resource_class._predicate  # type: ignore[attr-defined]
        self.assertEqual(
            cls_predicate.error_template,
            "{user} cannot pred {resource}",
        )

        self.assertEqual(
            resource._predicate.error_template,  # type: ignore[attr-defined]
            "{user} cannot pred {resource}",
        )

    def test_format_permission_check_error(self) -> None:
        """Test format_permission_check_error."""
        self.assertEqual(
            format_permission_check_error(
                self.scope.can_display, PermissionContext.for_ui_user(self.user)
            ),
            "playground cannot display scope debusine",
        )

    def test_enforce(self) -> None:
        """Test the enforce method."""
        context.set_permission_context_scoped(
            self.scope, PermissionContext.for_ui_anonymous()
        )

        with override_permission(Scope, "can_display", AllowAll):
            enforce(self.scope.can_display)
        with (
            override_permission(Scope, "can_display", DenyAll),
            self.assertRaisesRegex(
                PermissionDenied, "AnonymousUser cannot display scope debusine"
            ),
        ):
            enforce(self.scope.can_display)

    def predicate_testers(
        self,
        work_request: Collection[Allow],
        anonymous: Collection[Allow],
        pcs: Collection[PermissionContext],
    ) -> Generator[PredicateTester]:
        for w in work_request:
            for a in anonymous:
                for pc in pcs:
                    yield mock_pred_check(
                        self, work_request=w, anonymous=a, pc=pc
                    )
                    yield mock_pred_filter(
                        self, work_request=w, anonymous=a, pc=pc
                    )

    def test_predicates_worker_activation(self) -> None:
        token = self.playground.create_worker_activation_token()
        for tester in self.predicate_testers(
            work_request=(Allow.PASS, Allow.NEVER),
            anonymous=(Allow.PASS, Allow.NEVER),
            pcs=[PermissionContext.for_worker_activation(token)],
        ):
            with self.subTest(tester=repr(tester)):
                tester.assert_forced_false()
                tester.assert_passes_with_checks_disabled()

    def test_predicates_worker_token(self) -> None:
        token = self.playground.create_worker_token()
        for tester in self.predicate_testers(
            work_request=(Allow.PASS, Allow.NEVER),
            anonymous=(Allow.PASS, Allow.NEVER),
            pcs=[PermissionContext.for_worker(token)],
        ):
            with self.subTest(tester=repr(tester)):
                tester.assert_forced_false()
                tester.assert_passes_with_checks_disabled()

    def test_predicates_worker_pass(self) -> None:
        """Test predicates with worker, work_request=PASS."""
        for tester in self.predicate_testers(
            work_request=[Allow.PASS],
            anonymous=(Allow.PASS, Allow.NEVER),
            pcs=[self.work_request_pc],
        ):
            with self.subTest(tester=repr(tester)):
                tester.assert_called()
                tester.assert_passes_with_checks_disabled()

    def test_predicates_worker_never(self) -> None:
        """Test predicates with worker, work_request=NEVER."""
        for tester in self.predicate_testers(
            work_request=[Allow.NEVER],
            anonymous=(Allow.PASS, Allow.NEVER),
            pcs=[self.work_request_pc],
        ):
            with self.subTest(tester=repr(tester)):
                tester.assert_forced_false()
                tester.assert_passes_with_checks_disabled()

    def test_predicates_anonymous_never(self) -> None:
        """Test predicates with anonymous user, anonymous=NEVER."""
        for tester in self.predicate_testers(
            work_request=(Allow.PASS, Allow.NEVER),
            anonymous=[Allow.NEVER],
            pcs=[PermissionContext.for_ui_anonymous()],
        ):
            with self.subTest(tester=repr(tester)):
                tester.assert_forced_false()
                tester.assert_passes_with_checks_disabled()

    def test_predicates_anonymous_pass(self) -> None:
        """Test predicates with anonymous user, anonymous=PASS."""
        for tester in self.predicate_testers(
            work_request=[Allow.PASS, Allow.NEVER],
            anonymous=[Allow.PASS],
            pcs=[PermissionContext.for_ui_anonymous()],
        ):
            with self.subTest(tester=repr(tester)):
                tester.assert_called()
                tester.assert_passes_with_checks_disabled()

    def test_predicates_user(self) -> None:
        """Test predicates with explicitly given user."""
        for tester in self.predicate_testers(
            work_request=(Allow.PASS, Allow.NEVER),
            anonymous=(Allow.PASS, Allow.NEVER),
            pcs=[PermissionContext.for_ui_user(self.user)],
        ):
            with self.subTest(tester=repr(tester)):
                tester.assert_called()
                tester.assert_passes_with_checks_disabled()

    def test_get_resource_scope(self) -> None:
        """Test get_resource_scope."""
        workspace = self.playground.get_default_workspace()
        artifact, _ = self.playground.create_artifact(create_files=False)

        self.assertIsNone(get_resource_scope(self.user))
        self.assertEqual(get_resource_scope(self.group), self.scope)

        self.assertEqual(get_resource_scope(self.scope), self.scope)
        self.assertEqual(get_resource_scope(self.scope1), self.scope1)

        self.assertEqual(get_resource_scope(workspace), self.scope)
        self.assertEqual(get_resource_scope(artifact), self.scope)

    def test_get_resource_workspace(self) -> None:
        """Test get_resource_workspace."""
        workspace = self.playground.get_default_workspace()
        artifact, _ = self.playground.create_artifact(create_files=False)

        self.assertIsNone(get_resource_workspace(self.user))
        self.assertIsNone(get_resource_workspace(self.group))
        self.assertIsNone(get_resource_workspace(self.scope))

        self.assertEqual(get_resource_workspace(workspace), workspace)
        self.assertEqual(get_resource_workspace(artifact), workspace)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-admin command to sync user information with its sources of truth."""

from typing import Any, NamedTuple, NoReturn, override

import gitlab
from django.conf import settings
from django.core.management import CommandParser

from debusine.db.context import context
from debusine.db.models import Group, User
from debusine.db.models.auth import Identity
from debusine.django.management.debusine_base_command import DebusineBaseCommand
from debusine.server.signon.providers import GitlabProvider


class GroupCheckResults(NamedTuple):
    """Results of a group membership check."""

    external_group_name: str
    group: Group
    found: int
    unknown: int
    to_add: list[User]
    to_remove: list[User]


class GitlabCheck:
    """
    Account sync support for Gitlab providers.

    This requires the ``sync_token`` key to be set in the provider's
    ``options``, to be a gitlab API token that is able to access the membership
    information of the groups to be synced.
    """

    def __init__(self, provider: GitlabProvider) -> None:
        """Instantiate a GitlabCheck for a provider."""
        self.provider = provider
        sync_token = self.provider.options.get("sync_token")
        if sync_token is None:
            raise ValueError(
                f"Provider {provider.name} has no 'sync_token'"
                " configured in its options"
            )
        self.gl = gitlab.Gitlab(
            self.provider.url_issuer, private_tokeen=sync_token
        )
        self.identities = {
            i.subject: i
            for i in Identity.objects.filter(
                issuer=self.provider.name, user__isnull=False
            ).select_related("user")
        }

    def sync_group(
        self, salsa_group_name: str, debusine_group: Group
    ) -> GroupCheckResults:
        """Sync accounts for Gitlab groups."""
        salsa_group = self.gl.groups.get(id=salsa_group_name)
        current_debusine_members = {u.id: u for u in debusine_group.users.all()}

        found = 0
        unknown = 0
        to_add: list[User] = []
        for member in salsa_group.members_all.list(iterator=True):
            subject = str(member.id)
            identity = self.identities.get(subject, None)
            if identity is None:
                # User is not in our database
                unknown += 1
                continue
            assert identity.user is not None
            if current_debusine_members.pop(identity.user.id, None):
                found += 1
            else:
                to_add.append(identity.user)

        return GroupCheckResults(
            external_group_name=salsa_group_name,
            group=debusine_group,
            found=found,
            unknown=unknown,
            to_add=to_add,
            to_remove=list(current_debusine_members.values()),
        )


class Command(DebusineBaseCommand):
    """Check group membership against remote signon providers."""

    help = "Update group status based on membership changes on remote servers."

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            "--fix",
            action="store_true",
            help="Perform membership changes",
        )

    def apply_group_check_result(
        self, result: GroupCheckResults, dry_run: bool
    ) -> None:
        """Apply changes from a :py:class:`GroupCheckResults`."""
        self.stdout.write(
            f"{result.external_group_name} → {result.group}:"
            f" {result.found} valid members,"
            f" {len(result.to_add)} members to add,"
            f" {len(result.to_remove)} members to remove,"
            f" {result.unknown} users not found in Debusine"
        )
        if dry_run:
            for user in result.to_add:
                self.stdout.write(f"{user} should be added to {result.group}")
            for user in result.to_remove:
                self.stdout.write(
                    f"{user} should be removed from {result.group}"
                )
        else:
            with context.management_command():
                for user in result.to_add:
                    result.group.add_user(user)
                    self.stdout.write(f"{user} added to {result.group}")
                for user in result.to_remove:
                    result.group.remove_user(user)
                    self.stdout.write(f"{user} removed from {result.group}")

    def handle_gitlab(self, provider: GitlabProvider, *, dry_run: bool) -> None:
        """Handle account sync for a gitlab provider."""
        if not (add_to_group := provider.options.get("add_to_group")):
            self.stderr.write(
                f"{provider.name}: provider without add_to_group: skipping sync"
            )
            return

        try:
            gitlab_check = GitlabCheck(provider)
        except Exception as e:
            self.stderr.write(f"{provider.name}: {e}")
            return

        for external_group, debusine_group in add_to_group.items():
            if external_group.startswith("nm:"):
                self.stderr.write(
                    f"{external_group}: sync with nm.debian.org"
                    " is not yet implemented"
                )
            else:
                group = Group.objects.get_by_name(debusine_group)
                result = gitlab_check.sync_group(external_group, group)
                self.apply_group_check_result(result, dry_run)

    @override
    def handle(self, *args: Any, fix: bool = False, **options: Any) -> NoReturn:
        for provider in getattr(settings, "SIGNON_PROVIDERS", ()):
            match provider:
                case GitlabProvider():
                    self.handle_gitlab(provider, dry_run=not fix)
                case _:
                    self.stderr.write(
                        f"{provider.name}: provider unsupported for sync"
                    )

        raise SystemExit(0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command create_workflow."""

import io
from typing import Any, ClassVar, override

import yaml
from django.core.management import CommandError

from debusine.artifacts.models import TaskTypes
from debusine.client.models import RuntimeParameter
from debusine.db.models import (
    Artifact,
    User,
    WorkRequest,
    WorkflowTemplate,
    Workspace,
    default_workspace,
    system_user,
)
from debusine.django.management.tests import call_command
from debusine.test.django import TestCase


class CreateWorkflowCommandTests(TestCase):
    """Tests for the create_workflow command."""

    artifact: ClassVar[Artifact]
    template: ClassVar[WorkflowTemplate]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.artifact = cls.playground.create_source_artifact()
        architectures = ["amd64", "arm64"]
        for architecture in architectures:
            cls.playground.create_debian_environment(
                codename="bookworm", architecture=architecture
            )
        cls.template = WorkflowTemplate.objects.create(
            name="test",
            workspace=cls.playground.get_default_workspace(),
            task_name="sbuild",
            static_parameters={
                "input": {"source_artifact": cls.artifact.id},
                "architectures": architectures,
            },
            runtime_parameters=RuntimeParameter.ANY,
        )

    def assert_workflow(
        self,
        workflow: WorkRequest,
        *,
        workspace: Workspace | None = None,
        created_by: User | None = None,
        runtime_parameters: dict[str, Any] | None = None,
    ) -> None:
        self.assertEqual(workflow.workspace, workspace or default_workspace())
        self.assertEqual(workflow.created_by, created_by or system_user())
        self.assertEqual(workflow.status, WorkRequest.Statuses.PENDING)
        assert workflow.workflow_template is not None
        self.assertEqual(
            workflow.task_data,
            {
                **(runtime_parameters or {}),
                **workflow.workflow_template.static_parameters,
            },
        )

    def test_data_from_file(self) -> None:
        """`create_workflow` accepts data from a file."""
        data = {"target_distribution": "debian:bookworm"}
        data_file = self.create_temporary_file(
            contents=yaml.safe_dump(data).encode()
        )
        stdout, stderr, exit_code = call_command(
            "create_workflow", "test", "--data", str(data_file)
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        workflow = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        )
        self.assert_workflow(workflow, runtime_parameters=data)

    def test_data_from_stdin(self) -> None:
        """`create_workflow` accepts data from stdin."""
        data = {"target_distribution": "debian:bookworm"}
        stdout, stderr, exit_code = call_command(
            "create_workflow", "test", stdin=io.StringIO(yaml.safe_dump(data))
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        workflow = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        )
        self.assert_workflow(workflow, runtime_parameters=data)

    def test_empty_data(self) -> None:
        """`create_workflow` defaults data to {}."""
        WorkflowTemplate.objects.create(
            name="noop", workspace=default_workspace(), task_name="noop"
        )
        stdout, stderr, exit_code = call_command(
            "create_workflow", "noop", stdin=io.StringIO()
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        workflow = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW, task_name="noop"
        )
        self.assert_workflow(workflow, runtime_parameters={})

    def test_different_created_by(self) -> None:
        """`create_workflow` can use a different created-by user."""
        user = self.playground.get_default_user()
        data = {"target_distribution": "debian:bookworm"}
        stdout, stderr, exit_code = call_command(
            "create_workflow",
            "test",
            "--created-by",
            user.username,
            stdin=io.StringIO(yaml.safe_dump(data)),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        workflow = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        )
        self.assert_workflow(workflow, created_by=user, runtime_parameters=data)

    def test_different_workspace(self) -> None:
        """`create_workflow` can use a non-default workspace."""
        workspace_name = "test-workspace"
        workspace = self.playground.create_workspace(
            name=workspace_name, public=True
        )
        workspace.set_inheritance([self.playground.get_default_workspace()])
        self.template.workspace = workspace
        self.template.save()
        data = {"target_distribution": "debian:bookworm"}
        stdout, stderr, exit_code = call_command(
            "create_workflow",
            "test",
            "--workspace",
            workspace_name,
            stdin=io.StringIO(yaml.safe_dump(data)),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        workflow = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        )
        self.assert_workflow(
            workflow, workspace=workspace, runtime_parameters=data
        )

    def test_invalid_data_yaml(self) -> None:
        """`create_workflow` returns error: cannot parse data."""
        with self.assertRaisesRegex(
            CommandError, r"^Error parsing YAML:"
        ) as exc:
            call_command("create_workflow", "test", stdin=io.StringIO(":"))

        self.assertEqual(exc.exception.returncode, 3)

    def test_user_not_found(self) -> None:
        """`create_workflow` returns error: user not found."""
        with self.assertRaisesRegex(
            CommandError, r'^User "nonexistent" not found'
        ) as exc:
            call_command(
                "create_workflow",
                "test",
                "--created-by",
                "nonexistent",
                stdin=io.StringIO(),
            )

        self.assertEqual(exc.exception.returncode, 3)

    def test_workspace_not_found(self) -> None:
        """`create_workflow` returns error: workspace not found."""
        with self.assertRaisesRegex(
            CommandError, r"Workspace 'nonexistent' not found"
        ) as exc:
            call_command(
                "create_workflow",
                "test",
                "--workspace",
                "nonexistent",
                stdin=io.StringIO(),
            )

        self.assertEqual(exc.exception.returncode, 3)

    def test_template_name_not_found(self) -> None:
        """`create_workflow` returns error: template name not found."""
        with self.assertRaisesRegex(
            CommandError, r'^Workflow template "nonexistent" not found'
        ) as exc:
            call_command("create_workflow", "nonexistent", stdin=io.StringIO())

        self.assertEqual(exc.exception.returncode, 3)

    def test_bad_workflow_data(self) -> None:
        """`create_workflow` returns error: bad workflow data."""
        data = {"target_distribution": "debian:bookworm", "foo": "bar"}
        with self.assertRaisesRegex(
            CommandError,
            r"Extra inputs are not permitted",
        ) as exc:
            call_command(
                "create_workflow",
                "test",
                stdin=io.StringIO(yaml.safe_dump(data)),
            )

        self.assertIn("'loc': ('foo',)", str(exc.exception))

        self.assertEqual(exc.exception.returncode, 3)
        self.assertFalse(
            WorkRequest.objects.filter(
                task_type=TaskTypes.WORKFLOW, task_name="sbuild"
            ).exists()
        )
